"""
CommandLine:
    python ~/code/watch/dev/validate_annotation_schemas.py

Prereq:


    TEST_DPATH=$HOME/tmp/test_smart_schema
    mkdir -p $TEST_DPATH
    cd $TEST_DPATH
    git clone git@smartgitlab.com:TE/annotations.git
    git clone git@smartgitlab.com:infrastructure/docs.git

    pip install jsonschema ubelt -U


Also:

    python -m watch.demo.metrics_demo.generate_demodata --reset
    python ~/code/watch/dev/validate_annotation_schemas.py


References:
    https://smartgitlab.com/TE/annotations/-/issues/17
    https://smartgitlab.com/TE/standards/-/snippets/18


Example:

    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)

    python -m watch.cli.validate_annotation_schemas \
        --site_model_dpath="$DVC_DATA_DPATH"/annotations/drop6/site_models \
        --region_model_dpath="$DVC_DATA_DPATH"/annotations/drop6/region_models

"""

import ubelt as ub
import jsonschema
import json
import scriptconfig as scfg


class ValidateAnnotationConfig(scfg.DataConfig):
    """
    Validate the site / region model schemas
    """

    site_model_dpath = scfg.Value('auto', help='path to the site model directory')

    region_model_dpath = scfg.Value('auto', help='path to the region model directory')

    site_schema = scfg.Value('auto', help='path to the site model directory')

    region_schema = scfg.Value('auto', help='path to the region model directory')


def main():
    # Point to the cloned repos
    # docs_dpath = ub.Path('docs')
    # annotations_dpath = dvc_dpath / 'annotations/drop6'
    config = ValidateAnnotationConfig.cli()
    # config['site_model_dpath']

    # Load schemas
    # site_model_schema_fpath = docs_dpath / ('pages/schemas/site-model.schema.json')
    # region_model_schema_fpath = docs_dpath / ('pages/schemas/region-model.schema.json')
    # site_model_schema = json.loads(site_model_schema_fpath.read_text())
    # region_model_schema = json.loads(region_model_schema_fpath.read_text())

    if config['site_schema'] == 'auto':
        import watch
        site_schema = site_model_schema = watch.rc.registry.load_site_model_schema()
    else:
        raise NotImplementedError

    if config['region_schema'] == 'auto':
        import watch
        region_schema = region_model_schema = watch.rc.registry.load_region_model_schema()
    else:
        raise NotImplementedError

    if config['site_model_dpath'] == 'auto':
        import watch
        dvc_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        site_model_dpath = dvc_dpath / 'annotations/drop6/site_models'
    else:
        site_model_dpath = ub.Path(config['site_model_dpath'])

    if config['region_model_dpath'] == 'auto':
        import watch
        dvc_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        region_model_dpath = dvc_dpath / 'annotations/drop6/region_models'
    else:
        region_model_dpath = ub.Path(config['region_model_dpath'])

    site_model_fpaths = list(site_model_dpath.glob('*.geojson'))
    region_model_fpaths = list(region_model_dpath.glob('*.geojson'))

    #  hack to remove known invalid site models
    if 0:
        site_model_fpaths = [s for s in site_model_fpaths if 'Rxxx' not in s.stem]

    print('Validate Data Content')
    validate_data_contents(region_model_fpaths, site_model_fpaths)

    print('Validate Schemas')
    validate_schemas(site_model_fpaths, region_model_fpaths, site_model_schema,
                     region_model_schema)

    print('Sam Script')
    sam_script(site_model_fpaths, region_model_fpaths, site_schema, region_schema)


def sam_script(site_model_fpaths, region_model_fpaths, site_schema, region_schema):
    for test_file in site_model_fpaths + region_model_fpaths:
        try:
            model = json.load(ub.Path(test_file).open())
        except FileNotFoundError:
            print("Can't find the file {:s}".format(test_file))
            continue
        except json.decoder.JSONDecodeError:
            print("Can't load the JSON content of file {:s}".format(test_file))
            continue

        try:
            feature0 = model['features'][0]
        except IndexError:
            print('No features in the file {:s}'.format(test_file))
            continue

        try:
            feature_properties = feature0['properties']
        except KeyError:
            print('Properties missing from the first feature of {:s}'.format(test_file))
            continue

        try:
            feature_type = feature_properties['type']
        except KeyError:
            print('No feature type specified in the first feature of {:s}'.format(test_file))
            continue

        try:
            if feature_type in ['region', 'site_summary']:
                jsonschema.validate(instance=model, schema=region_schema)
            elif feature_type in ['site', 'observation']:
                jsonschema.validate(instance=model, schema=site_schema)
            else:
                raise ValueError(feature_type)
        except jsonschema.ValidationError as e:
            print('Validation failed for {}'.format(test_file))
            print(e)
            continue
        except ValueError:
            print('No type inferrable from the feature type {:s}'.format(feature_type))
            continue

        print('{} is valid'.format(test_file))


def validate_schemas(site_model_fpaths, region_model_fpaths, site_model_schema,
                     region_model_schema):

    region_errors = []
    prog = ub.ProgIter(region_model_fpaths, desc='check region models')
    for region_model_fpath in prog:
        region_model = json.loads(region_model_fpath.read_text())
        try:
            jsonschema.validate(region_model, region_model_schema)
        except jsonschema.ValidationError as ex:
            error_info = {
                'type': 'region_model_error',
                'name': region_model_fpath.stem,
                'ex': ex,
            }
            region_errors.append(error_info)
            prog.set_description(f'check region models, errors: {len(region_errors)}')

    site_errors = []
    prog = ub.ProgIter(site_model_fpaths, desc='check site models')
    for site_model_fpath in prog:
        site_model = json.loads(site_model_fpath.read_text())
        try:
            jsonschema.validate(site_model, site_model_schema)
        except jsonschema.ValidationError as ex:
            error_info = {
                'type': 'site_model_error',
                'name': site_model_fpath.stem,
                'ex': ex,
            }
            site_errors.append(error_info)
            prog.set_description(f'check site models, errors: {len(site_errors)}')

    print('site_errors = {}'.format(ub.urepr(site_errors, nl=1)))
    print('region_errors = {}'.format(ub.urepr(region_errors, nl=1)))

    print(f'{len(site_errors)} / {len(site_model_fpaths)} site model errors')
    print(f'{len(region_errors)} / {len(region_model_fpaths)} region model errors')


def validate_site_dataframe(site_df):
    from dateutil.parser import parse
    import numpy as np
    dummy_start_date = '1970-01-01'  # hack, could be more robust here
    dummy_end_date = '2101-01-01'
    first = site_df.iloc[0]
    rest = site_df.iloc[1:]
    assert first['type'] == 'site', 'first row must have type of site'
    assert first['region_id'] is not None, 'first row must have a region id'
    assert rest['type'].apply(lambda x: x == 'observation').all(), (
        'rest of row must have type observation')
    assert rest['region_id'].apply(lambda x: x is None).all(), (
        'rest of row must have region_id=None')

    site_start_date = first['start_date'] or dummy_start_date
    site_end_date = first['end_date'] or dummy_end_date
    site_start_datetime = parse(site_start_date)
    site_end_datetime = parse(site_end_date)

    if site_end_datetime < site_start_datetime:
        print('\n\nBAD SITE DATES:')
        print(first)

    # Check datetime errors in observations
    try:
        obs_dates = [None if x is None else parse(x) for x in rest['observation_date']]
        obs_isvalid = [x is None for x in obs_dates]
        valid_obs_dates = list(ub.compress(obs_dates, obs_isvalid))
        if not all(valid_obs_dates):
            # null_obs_sites.append(first[['site_id', 'status']].to_dict())
            pass
        valid_deltas = np.array([d.total_seconds() for d in np.diff(valid_obs_dates)])
        assert (valid_deltas >= 0).all(), 'observations must be sorted temporally'
    except AssertionError as ex:
        print('ex = {!r}'.format(ex))
        print(site_df)
        raise


def validate_region_model_content(region_df):
    from dateutil.parser import parse
    import os
    is_region = region_df['type'] == 'region'
    region_part = region_df[is_region]
    assert len(region_part) == 1, 'must have exactly one region in each region file'
    assert region_part['region_id'].apply(lambda x: x is not None).all(), 'regions must have region ids'

    region_rel_fpath = region_df.fpath.relative_to(region_df.fpath.parent.parent.parent)

    region_row = region_part.iloc[0]
    region_id = region_row['region_id']
    region_stem = region_df.fpath.stem

    errors = []
    region_report = {
        'region_id': region_id,
        'rel_fpath': os.fspath(region_rel_fpath),
        'errors': errors,
    }

    if region_id != region_stem:
        errors.append({
            'description': 'A region file name does not match its region id',
            'offending_info': {
                'region_id': region_id,
                'filename': region_df.fpath.name,
            }
        })

    dummy_start_date = '1970-01-01'  # hack, could be more robust here
    dummy_end_date = '2101-01-01'

    # Hack to set all region-ids
    sites_part = region_df[~is_region]
    assert (sites_part['type'] == 'site_summary').all(), 'rest of data must be site summaries'
    assert sites_part['region_id'].apply(lambda x: x is None).all(), ('site-summaries do not have region ids')

    region_start_date = region_row['start_date'] or dummy_start_date
    region_end_date = region_row['end_date'] or dummy_end_date

    region_start_datetime = parse(region_start_date)
    region_end_datetime = parse(region_end_date)
    if region_end_datetime < region_start_datetime:
        errors.append(f'Bad region dates: {region_start_datetime=}, {region_end_datetime=}')

    # Check datetime errors
    sitesum_start_dates = sites_part['start_date'].apply(lambda x: parse(x or region_start_date))
    sitesum_end_dates = sites_part['end_date'].apply(lambda x: parse(x or region_end_date))
    has_bad_time_range = sitesum_start_dates > sitesum_end_dates

    bad_date_rows = sites_part[has_bad_time_range]
    if len(bad_date_rows):
        bad_row_info = bad_date_rows[['site_id', 'start_date', 'end_date', 'originator']].to_dict('records')
        errors.append({
            'description': 'Site summary rows with start_dates > end_date',
            'offending_rows': bad_row_info,
        })
    if not region_report['errors']:
        region_report['errors'] = None
    region_report['num_errors'] = len(errors)
    return region_report


def validate_site_content(site_df):
    from dateutil.parser import parse
    import os
    dummy_start_date = '1970-01-01'  # hack, could be more robust here
    dummy_end_date = '2101-01-01'
    first = site_df.iloc[0]
    rest = site_df.iloc[1:]
    assert first['type'] == 'site', 'first row must have type of site'
    assert first['region_id'] is not None, 'first row must have a region id'
    assert rest['type'].apply(lambda x: x == 'observation').all(), (
        'rest of row must have type observation')
    assert rest['region_id'].apply(lambda x: x is None).all(), (
        'rest of row must have region_id=None')

    site_start_date = first['start_date'] or dummy_start_date
    site_end_date = first['end_date'] or dummy_end_date
    site_start_datetime = parse(site_start_date)
    site_end_datetime = parse(site_end_date)

    rel_fpath = site_df.fpath.relative_to(site_df.fpath.parent.parent.parent)

    errors = []
    site_report = {
        'region_id': first['region_id'],
        'site_id': first['site_id'],
        'rel_fpath': os.fspath(rel_fpath),
        'errors': errors,
    }

    if site_end_datetime < site_start_datetime:
        offending = first[['site_id', 'start_date', 'end_date', 'originator']].to_dict()
        errors.append({
            'description': 'Site summary row has a start_date > end_date',
            'offending_info': offending,
        })

    # Check datetime errors in observations
    null_obs_sites = []
    import numpy as np
    try:
        obs_dates = [None if x is None else parse(x) for x in rest['observation_date']]
        obs_isvalid = [x is not None for x in obs_dates]
        valid_obs_dates = list(ub.compress(obs_dates, obs_isvalid))
        if not all(valid_obs_dates):
            null_obs_sites.append(first[['site_id', 'status']].to_dict())
        valid_deltas = np.array([d.total_seconds() for d in np.diff(valid_obs_dates)])
        assert (valid_deltas >= 0).all(), 'observations must be sorted temporally'
    except AssertionError as ex:
        print('ex = {!r}'.format(ex))
        print(site_df)
        raise

    if null_obs_sites:
        errors.append({
            'description': 'sites contained null observations',
            'offending_info': null_obs_sites,
        })

    if not site_report['errors']:
        site_report['errors'] = None
    site_report['num_errors'] = len(errors)

    return site_report


def validate_data_contents(region_model_fpaths, site_model_fpaths):
    import geopandas as gpd
    # Start reading sites while we are doing other work
    site_read_pool = ub.JobPool(mode='process', max_workers=8)
    for site_model_fpath in ub.ProgIter(site_model_fpaths, desc='load site models'):
        job = site_read_pool.submit(gpd.read_file, site_model_fpath)
        job.fpath = site_model_fpath

    # Construct the region reports
    region_read_pool = ub.JobPool(mode='thread', max_workers=0)
    for region_model_fpath in ub.ProgIter(region_model_fpaths, desc='load region models'):
        job = region_read_pool.submit(gpd.read_file, region_model_fpath)
        job.fpath = region_model_fpath

    # Finish reading region models and build reports
    region_models = []
    region_reports = []
    for job in region_read_pool.as_completed(desc='collect region models'):
        region_df = job.result()
        region_df.fpath = job.fpath
        region_report = validate_region_model_content(region_df)
        region_models.append(region_df)
        region_reports.append(region_report)
    region_reports = sorted(region_reports, key=lambda x: (x['num_errors'], x['region_id']))
    print('region_reports = {}'.format(ub.urepr(region_reports, sort=0, nl=-1)).replace('\'', '"').replace('None', 'null'))

    # Finish reading site models, build reports while this is happening.
    site_reports = []
    site_models = []
    for job in site_read_pool.as_completed(desc='collect site models'):
        site_df = job.result()
        site_df.fpath = job.fpath
        site_report = validate_site_content(site_df)
        site_reports.append(site_report)
        site_models.append(site_df)

    region_id_to_site_reports = ub.group_items(site_reports, lambda x: x['region_id'])
    region_id_to_report_group = {}
    for region_id, sub_reports in region_id_to_site_reports.items():
        total_errors = sum([r['num_errors'] for r in sub_reports])
        invalid_site_reports = [r for r in sub_reports if r['num_errors']]
        for r in invalid_site_reports:
            r.pop('region_id', None)
        region_id_to_report_group[region_id] = {
            'region_id': region_id,
            'num_errors': total_errors,
            'valid_sites': len(sub_reports) - len(invalid_site_reports),
            'invalid_site_reports': invalid_site_reports,
        }

    grouped_site_reports = sorted(region_id_to_report_group.values(), key=lambda x: (x['num_errors'], x['region_id']))
    print('site_reports = {}'.format(ub.urepr(grouped_site_reports, sort=0, nl=-1)).replace('\'', '"').replace('None', 'null'))


if __name__ == '__main__':
    main()



"""
Utilities to configure / inspect system resources
"""


def request_nofile_limits(requested_limit='auto'):
    """
    Attempts to increase the limit of open file descriptors on the system.

    Args:
        requested_limit (int | str):
            attempt to increase rlimit_nofile to this number of open files,
            if auto, uses a hueristic

    Ignore:
        # Helpful file descriptor monitor script:
        watch -x bash -c '
            PROC_ID_LIST=($(ps -a | grep python | awk '"'"'{print $1}'"'"' ))
            for PROC_ID in "${PROC_ID_LIST[@]}"; do
                NUM_OPEN_FILES=$(lsof -p $PROC_ID | wc -l)
                echo "PROC_ID=$PROC_ID, NUM_OPEN_FILES=$NUM_OPEN_FILES"
            done
        '

        # Query current soft limit
        ulimit -S -n

        # Query current hard limit
        ulimit -H -n

        # Request higher soft limit
        ulimit -S -n 8192

    Example:
        >>> from watch.utils import util_resources
        >>> util_resources.request_nofile_limits()
    """
    import ubelt as ub
    if ub.LINUX:
        import resource
        if isinstance(requested_limit, str):
            if requested_limit == 'auto':
                requested_limit = 8192
            else:
                raise KeyError(requested_limit)
        soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
        if requested_limit > soft:
            print('Requesting FileLimit = {!r}'.format(requested_limit))
            print(' * Before FileLimit: soft={}, hard={}'.format(soft, hard))
            try:
                resource.setrlimit(resource.RLIMIT_NOFILE, (requested_limit, hard))
            except Exception as ex:
                print(f'ERROR ex={ex}')
            soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
            print(' * After FileLimit: soft={}, hard={}'.format(soft, hard))


def check_shm_limits(threshold_gibibytes=1):
    """
    When running training in docker you may need more shared memory than
    its default provide. Warns if shared memory is not above a threshold.

    Notes:
        Running docker run with `--shm-size=32768m` increases shm to 32GB

    Raises:
        IOError: if the system does not have a /dev/shm device.

    References:
        https://stackoverflow.com/questions/30210362/how-to-increase-the-size-of-the-dev-shm-in-docker-container
    """
    import ubelt as ub
    import psutil

    shm = ub.Path('/dev/shm')
    if shm.exists():
        usage = psutil.disk_usage(shm)
        shm_bytes = usage.total
    else:
        raise IOError('Dont know how to check shm size on this system')

    # shm_gigabytes = (shm_bytes / 10 ** 9)
    shm_gibibytes = (shm_bytes / 2 ** 30)

    if shm_gibibytes < threshold_gibibytes:
        import warnings
        print(f'shm_gibibytes={shm_gibibytes}')
        warnings.warn('Likely do not have enough /dev/shm space')

    if 0:
        disk_parts = psutil.disk_partitions(all=True)
        # Note sure if there is a way to find candidates that could be
        # /dev/shm or if it is ever the case that it doesnt exist.
        import pandas as pd
        rows = []
        for part in disk_parts:
            part_dict = {k: getattr(part, k) for k in dir(part) if not k.startswith('_')}
            part_dict = {k: v for k, v in part_dict.items() if isinstance(v, (str, float, int))}
            part_dict['opts'] = part_dict['opts'][0:128]
            rows.append(part_dict)
        df = pd.DataFrame(rows)
        df = df.sort_values('fstype')
        print(df.to_string())



import torchmetrics

import warnings
warnings.warn('''
    util_torchmetrics is deprecated since torchmetrics==0.10.0.
    Consider replacing calls to `util_torchmetrics.Binary{Metric}(...)`
    with calls to `torchmetrics.{Metric}(..., task='binary')`.
''', DeprecationWarning, stacklevel=2)

'''
This is a stopgap measure until https://github.com/Lightning-AI/metrics/pull/1195 is integrated
integrated into torchmetrics (targeted for v0.10?). The main issue that these metrics solve is
that when computing classification stats for *binary* problems, the returned score is an average
over the two classes stats when most of the time we only want one of them. This is especially
necessary in imbalanced settings where the metric for the largest class completely obscures the
impact of the smaller class.

SeeAlso:
    ../monkey/monkey_torchmetrics.py
'''


class BinaryF1Score(torchmetrics.StatScores):
    def compute(self):
        tp, fp, tn, fn = self._get_final_stats()
        tp, fp, tn, fn = tp[1], fp[1], tn[1], fn[1]
        return (2 * tp) / ((2 * tp) + fp + fn)


class BinaryOverallAccuracy(torchmetrics.StatScores):
    def compute(self):
        tp, fp, tn, fn = self._get_final_stats()
        tp, fp, tn, fn = tp[1], fp[1], tn[1], fn[1]
        return (tp + tn) / (tp + fp + tn + fn)


class BinaryBalancedAccuracy(torchmetrics.StatScores):
    def compute(self):
        tp, fp, tn, fn = self._get_final_stats()
        tp, fp, tn, fn = tp[1], fp[1], tn[1], fn[1]

        tpr = tp / (tp + fn)
        tnr = tn / (tn + fp)

        return (tpr + tnr) / 2


class BinaryTruePositiveRate(torchmetrics.StatScores):
    def compute(self):
        tp, fp, tn, fn = self._get_final_stats()
        tp, fp, tn, fn = tp[1], fp[1], tn[1], fn[1]

        return tp / (tp + fn)



"""

import sys
sys.path.append('/home/joncrall/code/SITS-Former/code')
from model import classification_model as clf

import liberator
lib = liberator.Liberator()
lib.add_dynamic(clf.BERTClassification)
lib.expand(['model'])
print(lib.current_sourcecode())

"""
from torch.nn.modules import LayerNorm
from torch.nn.modules.transformer import TransformerEncoder
from torch.nn.modules.transformer import TransformerEncoderLayer
import math
import torch.nn as nn
import torch


class PositionalEncoding(nn.Module):

    def __init__(self, d_model, max_len=366):
        super().__init__()

        # Compute the positional encodings once in log space.
        pe = torch.zeros(max_len + 1, d_model).float()
        pe.require_grad = False

        position = torch.arange(0, max_len).float(
        ).unsqueeze(1)         # [max_len, 1]
        div_term = (torch.arange(0, d_model, 2).float() * -
                    (math.log(10000.0) / d_model)).exp()  # [d_model/2,]

        # keep pe[0,:] to zeros
        # broadcasting to [max_len, d_model/2]
        pe[1:, 0::2] = torch.sin(position * div_term)
        # broadcasting to [max_len, d_model/2]
        pe[1:, 1::2] = torch.cos(position * div_term)

        self.register_buffer('pe', pe)

    def forward(self, time):
        output = torch.stack([torch.index_select(self.pe, 0, time[i, :])
                             for i in range(time.shape[0])], dim=0)
        return output       # [batch_size, seq_length, embed_dim]


class BERTEmbedding(nn.Module):
    """
    BERT Embedding which is consisted with under features
        1. InputEmbedding : project the input to embedding size through a lightweight 3D-CNN
        2. PositionalEncoding : adding positional information using sin/cos functions

        sum of both features are output of BERTEmbedding
    """

    def __init__(self, num_features, dropout=0.1):
        """
        :param num_features: number of input features
        :param dropout: dropout rate
        """
        super().__init__()
        channel_size = (32, 64, 256)
        kernel_size = (5, 3, 5, 3)

        self.conv1 = nn.Sequential(
            nn.Conv3d(in_channels=1,
                      out_channels=channel_size[0],
                      kernel_size=(kernel_size[0], kernel_size[1], kernel_size[1])),
            nn.ReLU(),
            nn.BatchNorm3d(channel_size[0]),
        )

        self.conv2 = nn.Sequential(
            nn.Conv3d(in_channels=channel_size[0],
                      out_channels=channel_size[1],
                      kernel_size=(kernel_size[2], kernel_size[3], kernel_size[3])),
            nn.ReLU(),
            nn.BatchNorm3d(channel_size[1]),
        )

        self.linear = nn.Linear(in_features=channel_size[1] * 2,
                                out_features=channel_size[2])

        self.embed_size = channel_size[-1]
        self.position = PositionalEncoding(
            d_model=self.embed_size, max_len=366)
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, input_sequence, doy_sequence):
        batch_size = input_sequence.size(0)
        seq_length = input_sequence.size(1)
        band_num = input_sequence.size(2)
        patch_size = input_sequence.size(3)
        first_dim = batch_size * seq_length

        obs_embed = input_sequence.view(
            first_dim, band_num, patch_size, patch_size).unsqueeze(1)
        obs_embed = self.conv1(obs_embed)
        obs_embed = self.conv2(obs_embed)
        # [batch_size*seq_length, embed_size]
        obs_embed = self.linear(obs_embed.view(first_dim, -1))
        obs_embed = obs_embed.view(batch_size, seq_length, -1)

        position_embed = self.position(doy_sequence)
        x = obs_embed + position_embed   # [batch_size, seq_length, embed_size]

        return self.dropout(x)


class BERT(nn.Module):

    def __init__(self, num_features, hidden,
                 n_layers, attn_heads, dropout=0.1):
        """
        :param num_features: number of input features
        :param hidden: hidden size of the SITS-Former model
        :param n_layers: numbers of Transformer blocks (layers)
        :param attn_heads: number of attention heads
        :param dropout: dropout rate
        """

        super().__init__()
        self.hidden = hidden
        self.n_layers = n_layers
        self.attn_heads = attn_heads

        feed_forward_hidden = hidden * 4

        self.embedding = BERTEmbedding(num_features)

        encoder_layer = TransformerEncoderLayer(
            hidden, attn_heads, feed_forward_hidden, dropout)
        encoder_norm = LayerNorm(hidden)

        self.transformer_encoder = TransformerEncoder(
            encoder_layer, n_layers, encoder_norm)

    def forward(self, x, doy, mask):
        mask = mask == 0

        x = self.embedding(input_sequence=x, doy_sequence=doy)

        x = x.transpose(0, 1)
        x = self.transformer_encoder(x, src_key_padding_mask=mask)
        x = x.transpose(0, 1)

        return x


class MulticlassClassification(nn.Module):

    def __init__(self, hidden, num_classes):
        super().__init__()
        self.linear = nn.Linear(hidden, num_classes)

    def forward(self, x, mask):
        mask = (1 - mask.unsqueeze(-1)) * 1e6
        x = x - mask        # mask invalid timesteps
        x, _ = torch.max(x, dim=1)      # max-pooling
        x = self.linear(x)
        return x


class BERTClassification(nn.Module):
    """
    Downstream task: Satellite Time Series Classification
    """

    def __init__(self, bert: BERT, num_classes):
        """
        :param bert: the BERT-Former model
        :param num_classes: number of classes to be classified
        """

        super().__init__()
        self.bert = bert
        self.classification = MulticlassClassification(
            self.bert.hidden, num_classes)

    def forward(self, x, doy, mask):
        x = self.bert(x, doy, mask)     # [batch_size, seq_length, embed_size]
        return self.classification(x, mask)



import pytorch_lightning as pl
import ubelt as ub
from typing import Dict, Any, Optional


class StateLogger(pl.callbacks.Callback):
    """
    Prints out what callbacks are being called

    DEPRECATE: Use text_logger
    """

    def __init__(self):
        pass

    def setup(self, trainer: 'pl.Trainer', pl_module: 'pl.LightningModule', stage: Optional[str] = None) -> None:
        print('setup state logger')
        print('trainer.default_root_dir = {!r}'.format(trainer.default_root_dir))

    def teardown(self, trainer: 'pl.Trainer', pl_module: 'pl.LightningModule', stage: Optional[str] = None) -> None:
        if 0:
            print('teardown state logger')

    if 0:
        def on_init_start(self, trainer: 'pl.Trainer') -> None:
            if 0:
                print('on_init_start')

        def on_init_end(self, trainer: 'pl.Trainer') -> None:
            if 0:
                print('on_init_start')

    def on_fit_start(self, trainer: 'pl.Trainer', pl_module: 'pl.LightningModule') -> None:
        if 0:
            print('on_fit_start')

    def on_fit_end(self, trainer: 'pl.Trainer', pl_module: 'pl.LightningModule') -> None:
        if 0:
            print('on_fit_end')

    # def on_train_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
    #     print('on_train_start')

    # def on_train_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
    #     print('on_train_end')

    def on_save_checkpoint(self, trainer: 'pl.Trainer', pl_module: 'pl.LightningModule', checkpoint: Dict[str, Any]) -> dict:
        if 0:
            print('on_save_checkpoint - checkpoint = {}'.format(ub.urepr(checkpoint.keys(), nl=1)))

    # def load_state_dict...
    # def on_load_checkpoint(self, trainer: 'pl.Trainer', pl_module: 'pl.LightningModule', callback_state: Dict[str, Any]) -> None:
    #     if 0:
    #         print('on_load_checkpoint - callback_state = {}'.format(ub.urepr(callback_state.keys(), nl=1)))

    def on_sanity_check_start(self, trainer: 'pl.Trainer', pl_module: 'pl.LightningModule') -> None:
        if 0:
            print('on_sanity_check_start')

    def on_sanity_check_end(self, trainer: 'pl.Trainer', pl_module: 'pl.LightningModule') -> None:
        if 0:
            print('on_sanity_check_end')

    def on_exception(self, trainer: 'pl.Trainer', pl_module: 'pl.LightningModule', *args, **kw) -> None:
        print('on_exception')
        print('kw = {!r}'.format(kw))
        print('args = {!r}'.format(args))
        print('INTERUPT')
        print('trainer.default_root_dir = {!r}'.format(trainer.default_root_dir))
        print('trainer.log_dir = {!r}'.format(trainer.log_dir))




def fix_sorted_set():
    # (kwcoco 0.5.1 renamed SortedSetQuiet to SortedSet)
    import kwcoco
    if('SortedSetQuiet' not in dir(kwcoco._helpers)  # noqa: E275
       and 'SortedSet' in dir(kwcoco._helpers)):
        kwcoco._helpers.SortedSetQuiet = kwcoco._helpers.SortedSet



"""
Example:
    >>> # xdoctest: +REQUIRES(env:HAS_DVC)
    >>> import numpy as np
    >>> import watch
    >>> from watch.tasks.depthPCD.model import getModel
    >>> #expt_dvc_dpath = watch.find_smart_dvc_dpath(tags='phase2_expt', hardware='auto')
    >>> model = getModel()
    >>> #model.load_weights(expt_dvc_dpath + '/models/depthPCD/basicModel.h5')
    >>> model.predict(np.zeros((1,400,400,3)))
"""



#!/usr/bin/env python3
r"""
Prediction script for Rutgers Material Semenatic Segmentation Models

CommandLine:

    export CUDA_VISIBLE_DEVICES=1
    DVC_DPATH=$(geowatch_dvc)

    KWCOCO_BUNDLE_DPATH=$DVC_DPATH/Drop2-Aligned-TA1-2022-02-15
    KWCOCO_BUNDLE_DPATH=$DVC_DPATH/Aligned-Drop3-TA1-2022-03-10
    BASE_COCO_FPATH=$KWCOCO_BUNDLE_DPATH/data.kwcoco.json

    RUTGERS_MATERIAL_MODEL_FPATH="$DVC_DPATH/models/rutgers/rutgers_peri_materials_v3/experiments_epoch_18_loss_59.014100193977356_valmF1_0.18694573888313187_valChangeF1_0.0_time_2022-02-01-01:53:20.pth"

    RUTGERS_MATERIAL_COCO_FPATH=$KWCOCO_BUNDLE_DPATH/rutgers_material_seg_v3.kwcoco.json

    # Generate Rutgers Features
    python -m watch.tasks.rutgers_material_seg.predict \
        --test_dataset=$BASE_COCO_FPATH \
        --checkpoint_fpath=$RUTGERS_MATERIAL_MODEL_FPATH  \
        --default_config_key=iarpa \
        --pred_dataset=$RUTGERS_MATERIAL_COCO_FPATH \
        --num_workers=avail \
        --batch_size=4 --devices auto:1
        --export_raw_features True

"""
import os
import torch
import datetime
import random
import kwcoco
import kwimage
import kwarray
import ndsampler
import numpy as np
from tqdm import tqdm  # NOQA
import ubelt as ub
import pathlib
import watch.tasks.rutgers_material_seg.utils.utils as utils
from watch.utils import util_parallel
from watch.tasks.rutgers_material_seg.models import build_model
from watch.tasks.rutgers_material_seg.datasets.iarpa_contrastive_dataset import SequenceDataset
import torch.nn.functional as F

try:
    from xdev import profile
except Exception:
    profile = ub.identity


class Evaluator(object):
    """
    CommandLine:
        DVC_DPATH=$(geowatch_dvc)
        DVC_DPATH=$DVC_DPATH xdoctest -m watch.tasks.rutgers_material_seg.predict Evaluator

    Example:
        >>> # xdoctest: +REQUIRES(env:DVC_DPATH)
        >>> from watch.tasks.rutgers_material_seg.predict import *  # NOQA
        >>> import watch
        >>> dvc_dpath = watch.find_smart_dvc_dpath()
        >>> checkpoint_fpath = dvc_dpath / 'models/rutgers/rutgers_peri_materials_v3/experiments_epoch_18_loss_59.014100193977356_valmF1_0.18694573888313187_valChangeF1_0.0_time_2022-02-01-01:53:20.pth'
        >>> #checkpoint_fpath = dvc_dpath / 'models/rutgers/experiments_epoch_62_loss_0.09470022770735186_valmIoU_0.5901660531463717_time_2021101T16277.pth'
        >>> #  Write out smaller version of the dataset
        >>> import kwcoco
        >>> dset = kwcoco.CocoDataset(dvc_dpath / 'Drop2-Aligned-TA1-2022-02-15/data_nowv_vali.kwcoco.json')
        >>> images = dset.videos(names=['KR_R001']).images[0]
        >>> sub_images = images.compress([s != 'WV' for s in images.lookup('sensor_coarse')])[::100]
        >>> sub_dset = dset.subset(sub_images)
        >>> sub_dset.fpath = (dvc_dpath / 'Drop2-Aligned-TA1-2022-02-15/small_test_data_nowv_vali.kwcoco.json')
        >>> sub_dset.dump(sub_dset.fpath)
        >>> input_kwcoco = sub_dset.fpath
        >>> #
        >>> src_coco_fpath = sub_dset.fpath
        >>> dst_coco_fpath = dvc_dpath / 'Drop2-Aligned-TA1-2022-02-15/mat_test.kwcoco.json'
        >>> kwargs = dict(
        >>>     default_config_key='iarpa',
        >>>     checkpoint_fpath=checkpoint_fpath,
        >>>     test_dataset=src_coco_fpath,
        >>>     pred_dataset=dst_coco_fpath,
        >>>     feat_dpath=dst_coco_fpath.parent / '_assets/test_rutgers_material_seg',
        >>>     devices='auto:1',
        >>>     num_workers='avail',
        >>>     save_raw_features=1,
        >>>     batch_size=32,
        >>>     cache=0,
        >>>     #num_workers=0,
        >>>     #gpu=None,
        >>> )
        >>> cmdline = False
        >>> evaler = build_evaler(cmdline=cmdline, **kwargs)
        >>> self = evaler
        >>> evaler.forward()
    """

    def __init__(self,
                 model: object,
                 checkpoint_fpath: str,
                 input_coco_dset: SequenceDataset,
                 output_coco_dataset: kwcoco.CocoDataset,
                 write_probs : bool = True,
                 device=None,
                 config : dict = None,
                 output_feat_dpath : pathlib.Path = None,
                 batch_size=1,
                 num_workers=0,
                 imwrite_kw={},
                 cache: bool = False,
                 save_raw_features: bool = False):
        """
        Evaluator class

        Args:
            model (object): trained or untrained model
            eval_loader (SequenceDataset): dataset with evaluation data
            optimizer (object): optimizer to train with
            scheduler (object): scheduler to train with
            cache (bool): if True, will only predict on images that do
                not have a feature computed for them yet.
        """

        self.model = model
        self.checkpoint_fpath = checkpoint_fpath
        self.save_raw_features = save_raw_features
        self.num_workers = num_workers
        self.output_coco_dataset = output_coco_dataset
        self.write_probs = write_probs
        self.device = device
        self.config = config
        self.num_classes = self.config['data']['num_classes']
        self.output_feat_dpath = output_feat_dpath
        self.stitcher_dict = {}
        if self.save_raw_features:
            # self.stitcher_dict_up3 = {}
            self.stitcher_dict_up5 = {}
        self.finalized_gids = set()
        self.imwrite_kw = imwrite_kw
        self.cache = cache
        self.batch_size = batch_size

        self.input_coco_dset = input_coco_dset

        # Hack together a channel code
        self.chan_code = '|'.join(['matseg_{}'.format(i) for i in range(self.num_classes)])
        # self.chan_code = '|'.join(['matseg.{}'.format(i) for i in range(self.num_classes)])

        self.output_channels = kwcoco.FusedChannelSpec.coerce(self.chan_code).concise()
        self.concise_chan_code = self.output_channels.spec
        self.concise_chan_path_code = self.output_channels.path_sanitize()

    def _output_path_for_image(self, gid):
        img = self.output_coco_dataset.index.imgs[gid]
        img_name = img.get('name', f'gid{gid:08d}')
        save_path = self.output_feat_dpath / f'{img_name}_{self.concise_chan_path_code}.tif'
        return save_path

    def _features_path_for_image(self, gid, layer):
        img = self.output_coco_dataset.index.imgs[gid]
        img_name = img.get('name', f'gid{gid:08d}')
        # FIXME: this shouldn't use the concise channel code, this should
        # use the same name as the path-sanatized channel string.
        save_path = self.output_feat_dpath / f'{img_name}_{self.concise_chan_path_code}_{layer}.tif'
        return save_path

    @profile
    def finalize_image(self, gid, cached=False):
        """
        Finializes the probabilities accumulated in the stitcher for a
        particular image, saves it to disk, and updates the kwcoco file in
        memory.

        Args:
            gid (int): image id to finalize
            cached (bool): if True, only updates the kwcoco file, assumes
                the file exists on disk.
        """
        img = self.output_coco_dataset.index.imgs[gid]

        self.finalized_gids.add(gid)

        if not cached:
            stitcher = self.stitcher_dict[gid]
            if self.save_raw_features:
                # stitcher_up3 = self.stitcher_dict_up3[gid]
                stitcher_up5 = self.stitcher_dict_up5[gid]

            import warnings
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', 'invalid value encountered in true_divide', category=RuntimeWarning)
                recon = stitcher.finalize()

                if self.save_raw_features:
                    # recon_up3 = stitcher_up3.finalize()
                    recon_up5 = stitcher_up5.finalize()
                    # self.stitcher_dict_up3.pop(gid)
                    self.stitcher_dict_up5.pop(gid)
            self.stitcher_dict.pop(gid)
        else:
            recon = None
            if self.save_raw_features:
                # recon_up3 = None
                recon_up5 = None

        from watch.tasks.fusion.predict import quantize_float01
        # Note using -11 and +11 as a tradeoff range because we cannot
        # guarentee the bounds of this data. Usually it is mean zero with
        # a std < 3, so this should be a decent range to work within.
        quant_recon, quantization = quantize_float01(recon, old_min=-11, old_max=11)
        nodata = quantization['nodata']

        save_path = self._output_path_for_image(gid)
        save_path = os.fspath(save_path)

        if self.save_raw_features:
            # quant_recon_up3, quantization_up3 = quantize_float01(recon_up3, old_min=-11, old_max=11)
            # nodata_up3 = quantization_up3['nodata']
            # save_path_up3 = self._features_path_for_image(gid, 'up3')
            # save_path_up3 = os.fspath(save_path_up3)

            quant_recon_up5, quantization_up5 = quantize_float01(recon_up5, old_min=-11, old_max=11)
            nodata_up5 = quantization_up5['nodata']
            save_path_up5 = self._features_path_for_image(gid, 'up5')
            save_path_up5 = os.fspath(save_path_up5)

        if cached:
            aux_height, aux_width = kwimage.load_image_shape(save_path)[0:2]

            if self.save_raw_features:
                # aux_height_up3, aux_width_up3 = kwimage.load_image_shape(save_path_up3)[0:2]
                aux_height_up5, aux_width_up5 = kwimage.load_image_shape(save_path_up5)[0:2]
        else:
            kwimage.imwrite(save_path, quant_recon, backend='gdal', space=None,
                            nodata=nodata, **self.imwrite_kw)
            aux_height, aux_width = quant_recon.shape[0:2]

            if self.save_raw_features:
                # kwimage.imwrite(save_path_up3, quant_recon_up3, backend='gdal', space=None,
                #                 nodata=nodata_up3, **self.imwrite_kw)
                # aux_height_up3, aux_width_up3 = quant_recon_up3.shape[0:2]

                kwimage.imwrite(save_path_up5, quant_recon_up5, backend='gdal', space=None,
                                nodata=nodata_up5, **self.imwrite_kw)
                aux_height_up5, aux_width_up5 = quant_recon_up5.shape[0:2]

        warp_aux_to_img = kwimage.Affine.scale(
            (img['width'] / aux_width,
             img['height'] / aux_height))

        aux = {
            'file_name': save_path,
            'height': aux_height,
            'width': aux_width,
            'channels': self.concise_chan_code,
            'warp_aux_to_img': warp_aux_to_img.concise(),
            'quantization': quantization,
        }

        auxiliary = img.setdefault('auxiliary', [])
        auxiliary.append(aux)

        if self.save_raw_features:
            # aux_up3 = {
            #     'file_name': save_path_up3,
            #     'height': aux_height,
            #     'width': aux_width,
            #     'channels': 'mat_up3:128',
            #     'warp_aux_to_img': warp_aux_to_img.concise(),
            #     'quantization': quantization_up3,
            # }
            # auxiliary.append(aux_up3)

            aux_up5 = {
                'file_name': save_path_up5,
                'height': aux_height,
                'width': aux_width,
                'channels': 'mat_up5:64',
                'warp_aux_to_img': warp_aux_to_img.concise(),
                'quantization': quantization_up5,
            }

            auxiliary.append(aux_up5)

    @profile
    def eval(self) -> tuple:
        """
        Execute predictions on the evaluator kwcoco file.

        Dumps images and the final kwcoco file to disk.
        """
        # from watch.utils import util_kwimage
        current_gids = []
        previous_gids = []

        config = self.config
        window_dims = (config['data']['time_steps'], config['data']['image_size'], config['data']['image_size'])  # [t,h,w]
        input_dims = (config['data']['image_size'], config['data']['image_size'])
        channels = config['data']['channels']

        if self.cache:
            # Detect if we already wrote predictions for some images.
            miss_gids = []
            hit_gids = []
            all_gids = list(self.input_coco_dset.index.imgs.keys())
            for gid in all_gids:

                fpaths = []
                fpath = self._output_path_for_image(gid)
                fpaths.append(fpath)
                if self.save_raw_features:
                    save_path_up5 = self._features_path_for_image(gid, 'up5')
                    # save_path_up3 = self._features_path_for_image(gid, 'up3')
                    fpaths.append(save_path_up5)
                    # fpaths.append(save_path_up3)

                if all(p.exists() for p in fpaths):
                    hit_gids.append(gid)
                else:
                    miss_gids.append(gid)
            # NOTE: WE ARE ASSUMING ONLY 1 TIME STEP IS USED
            # THIS HAS TO CHANGE IF MULTIPLE TIMESTEPS ARE USED
            subset_input_coco = self.input_coco_dset.subset(miss_gids)
            print(f'Found {len(hit_gids)} / {len(all_gids)} cached features')
        else:
            subset_input_coco = self.input_coco_dset

        # Build the torch datasets / loaders
        sampler = ndsampler.CocoSampler(subset_input_coco)
        window_overlap = 0.3  # TODO parametarize
        eval_dataset = SequenceDataset(
            sampler, window_dims, input_dims, channels, training=False,
            window_overlap=window_overlap, inference_only=True)
        print(f'{len(eval_dataset)=}')

        if len(eval_dataset) == 0:
            # hack for case where everything is already cached
            eval_dataloader = []
        else:
            eval_dataloader = eval_dataset.make_loader(
                batch_size=self.batch_size,
                num_workers=self.num_workers,
            )

        dataloader_iter = iter(eval_dataloader)
        writer = util_parallel.BlockingJobQueue(max_workers=self.num_workers)

        seen = set()

        if len(eval_dataloader):
            print('read self.checkpoint_fpath = {!r}'.format(self.checkpoint_fpath))
            checkpoint_state = torch.load(self.checkpoint_fpath)
            self.model.load_state_dict(checkpoint_state['model'])
            print(f"loadded model weights from: {self.checkpoint_fpath}")
            # print(f"Missing keys from loaded model: {missing_keys}, unexpected keys: {unexpexted_keys}")
            self.model = self.model.to(self.device)
            self.model.eval()

        with torch.no_grad():
            # from functools import partial
            # Prog = partial(ub.ProgIter, verbose=3)
            Prog = tqdm
            pbar = Prog(enumerate(dataloader_iter), total=len(eval_dataloader), desc='predict rutgers')
            for batch_index, batch in pbar:
                outputs = batch

                images = outputs['inputs']['im'].data[0]
                original_width, original_height = outputs['tr'].data[0][0]['space_dims']

                images = images.clone()

                bs, c, t, h, w = images.shape

                image1 = images[:, :, 0, :, :]
                image1 = image1.to(self.device)

                input_mask = image1.isnan()

                # Cannot input nan values to the model, so keep them as their
                # imputed value
                image1 = nan_normalize(
                    image1, dim=1, p=2,
                    imputation={'method': 'mean', 'dim': (0, 2, 3)},
                    keepna=False, mask=input_mask)

                output1, outputs_layers = self.model(image1)  # [B,22,150,150]

                # replace model outputs with nans in spatial locations
                output_mask = input_mask.any(dim=1, keepdims=True).expand_as(output1)
                output1[output_mask] = float('nan')

                # print(f"output: {output1.shape}, type: {output1.dtype}")

                bs, c, h, w = output1.shape
                output1_to_save = output1.permute(0, 2, 3, 1).cpu().detach().numpy()
                if self.save_raw_features:
                    # up3_to_save = outputs_layers['up3']
                    up5_to_save = outputs_layers['up5']

                    # up3_to_save = F.interpolate(up3_to_save, size=output1.shape[-2:], mode='bilinear', align_corners=False)
                    up5_to_save = F.interpolate(up5_to_save, size=output1.shape[-2:], mode='bilinear', align_corners=False)

                    # up3_to_save = up3_to_save.cpu().detach().numpy()
                    up5_to_save = up5_to_save.cpu().detach().numpy()
                    # bs, c_up3, h, w  = up3_to_save.shape
                    bs, c_up5, h, w  = up5_to_save.shape

                # print(f"output1_to_save: {output1_to_save.shape}")
                # print(f"output1_to_save min: {output1_to_save.min()}, max: {output1_to_save.max()}")
                # output_show = (output1_to_save - output1_to_save.min()) / (output1_to_save.max() - output1_to_save.min())
                # image_show1 = np.transpose(image1.cpu().detach().numpy()[0, :, :, :], (1, 2, 0))[:, :, :3]
                # image_show1 = (image_show1 - image_show1.min()) / (image_show1.max() - image_show1.min())
                # import matplotlib.pyplot as plt
                # fig = plt.figure()
                # ax1 = fig.add_subplot(1,2,1)
                # ax2 = fig.add_subplot(1,2,2)
                # ax1.imshow(image_show1)
                # ax2.imshow(output_show[0,:,:,:3])
                # plt.show()

                if self.write_probs:
                    for b in range(bs):
                        if len(current_gids) == 0:
                            current_gids = outputs['tr'].data[0][b]['gids']
                        else:
                            previous_gids = current_gids
                            current_gids = outputs['tr'].data[0][b]['gids']
                            mutually_exclusive = (set(previous_gids) - set(current_gids))
                            for gid in mutually_exclusive:
                                pbar.set_postfix_str('finalized {}'.format(len(self.finalized_gids)))
                                seen.add(gid)
                                writer.submit(self.finalize_image, gid)
                                # self.finalize_image(gid)

                        for gid in current_gids:
                            output = output1_to_save[b, :, :, :]

                            if self.save_raw_features:
                                # up3 = up3_to_save[b, :, :, :]
                                up5 = up5_to_save[b, :, :, :]

                                # up3 = np.transpose(up3, (1, 2, 0))
                                up5 = np.transpose(up5, (1, 2, 0))

                            if gid not in self.stitcher_dict.keys():
                                self.stitcher_dict[gid] = kwarray.Stitcher(
                                    (*outputs['tr'].data[0][b]['space_dims'],
                                     self.num_classes),
                                    device='numpy')

                                if self.save_raw_features:
                                    # self.stitcher_dict_up3[gid] = kwarray.Stitcher(
                                    #     (*outputs['tr'].data[0][b]['space_dims'], c_up3),
                                    #     device='numpy')
                                    self.stitcher_dict_up5[gid] = kwarray.Stitcher(
                                        (*outputs['tr'].data[0][b]['space_dims'], c_up5),
                                        device='numpy')
                            slice_ = outputs['tr'].data[0][b]['space_slice']

                            from watch.tasks.fusion.predict import CocoStitchingManager
                            CocoStitchingManager._stitcher_center_weighted_add(self.stitcher_dict[gid], slice_, output)

                            if self.save_raw_features:
                                # CocoStitchingManager._stitcher_center_weighted_add(self.stitcher_dict_up3[gid], slice_, up3)
                                CocoStitchingManager._stitcher_center_weighted_add(self.stitcher_dict_up5[gid], slice_, up5)

        writer.wait_until_finished()  # prevents a race condition

        if self.write_probs:
            # Finalize any remaining images
            for gid in Prog(list(self.stitcher_dict.keys()), desc='finish finalization'):
                # self.finalize_image(gid)
                writer.submit(self.finalize_image, gid)
                pbar.set_postfix_str('finalized {}'.format(len(self.finalized_gids)))

        writer.wait_until_finished()

        if self.cache:
            # Still need to update the kwcoco file for all of the hit gids
            for gid in Prog(hit_gids, desc='update coco file for cached images'):
                self.finalize_image(gid, cached=True)

        # export predictions to a new kwcoco file
        self.output_coco_dataset._invalidate_hashid()
        self.output_coco_dataset.dump(self.output_coco_dataset.fpath, newlines=True)

        return

    def forward(self) -> tuple:
        """forward pass for all epochs

        Args:
            cometml_experiment (object): comet ml experiment for logging
            world_size (int, optional): for distributed training. Defaults to 8.

        Returns:
            tuple: (train losses, validation losses, mIoU)
        """

        if self.config['procedures']['validate']:
            self.eval()
        return


def impute(a, imputation='zero', mask=None):
    """
    Replaces nan values according to a imputation method

    Args:
        a (Tensor): input data

        imputation (dict | str):
            dictionary containing keys:
                method (str): either zeros or mean

            if this is a string, it becomes the method in an imputation
            dictionary created with auto defaults.

        mask (Tensor):
            precomputed nan mask
    """
    if mask is None:
        mask = torch.isnan(a)

    if isinstance(imputation, str):
        imputation = {
            'method': imputation
        }
    imputation_method = imputation['method']
    if imputation_method == 'zero':
        out = torch.nan_to_num(a)
    elif imputation_method == 'mean':
        out = a.clone()
        mean_dims = imputation.get('dim', None)
        if mean_dims is None:
            mean = a.nanmean()
            out[mask] = mean
        else:
            fill_values = a.nanmean(dim=mean_dims, keepdims=True).expand_as(out)
            out[mask] = fill_values[mask]
    else:
        raise KeyError(imputation_method)
    return out


def nan_normalize(a, dim, p=2, imputation='zero', assume_nans=False,
                  keepna=True, mask=None):
    """
    Like torch.functional.normalize, but handles nans

    Args:
        a (Tensor): input data

        dim (int): dimension to normalize over

        p (int): type of norm

        imputation (dict | str):

            See :func:`impute`

            dictionary containing keys:
                method (str): either zeros or mean

            if this is a string, it becomes the method in an imputation
            dictionary created with auto defaults.

        assume_nans (bool):
            If true, skips the check if any nans exist and assume they do.
            Otherwise we check if there are no nans and just use normal
            normalize.

        keepna (bool):
            if False, keep the imputed results rather than re-masking them.

        mask (Tensor):
            if specified, use these as masked values

    Returns:
        Tensor: normalized array

    Example:
        >>> shape = (7, 5, 3)
        >>> a = data = torch.from_numpy(np.arange(np.prod(shape)).reshape(*shape)).float()
        >>> a[0:2, 0:2, 0:2] = float('nan')
        >>> dim = 2
        >>> p = 2
        >>> r1 = nan_normalize(a, dim, p, imputation='zero')
        >>> r2 = nan_normalize(a, dim, p, imputation='mean')
        >>> assert r1.isnan().sum() == a.isnan().sum()
        >>> assert r2.isnan().sum() == a.isnan().sum()

        >>> nan_data = torch.full((3, 2), fill_value=float('nan'))
        >>> dim = 1
        >>> nan_result = nan_normalize(nan_data, dim, p, imputation='mean')
        >>> assert torch.isnan(nan_result).all()

        >>> # Ensure this works when no nans exist
        >>> clean_data = torch.rand(3, 2)
        >>> v1 = nan_normalize(clean_data, dim, p, imputation='mean')
        >>> v2 = nan_normalize(clean_data, dim, p, imputation='mean', assume_nans=True)
    """
    if mask is None:
        mask = torch.isnan(a)
    if assume_nans or mask.any():
        if isinstance(imputation, str):
            imputation = {
                'method': imputation
            }
        # mean_dims = imputation.get('dim', 'auto')
        # if mean_dims == 'auto':
        #     mean_dims = tuple([i for i in range(len(a.shape)) if i != dim])
        out = impute(a, imputation=imputation, mask=mask)
        F.normalize(out, dim=dim, p=p, out=out)
        if keepna:
            out[mask] = float('nan')
    else:
        out = F.normalize(a, dim=dim, p=p)
    return out


def make_predict_config(cmdline=False, **kwargs):
    """
    Configuration for material prediction
    """
    from watch.utils import configargparse_ext
    from scriptconfig.smartcast import smartcast
    parser = configargparse_ext.ArgumentParser(
        add_config_file_help=False,
        description='Prediction script for the fusion task',
        auto_env_var_prefix='WATCH_RUTGERS_MATERIAL_PREDICT_',
        add_env_var_help=True,
        formatter_class='raw',
        config_file_parser_class='yaml',
        args_for_setting_config_path=['--config'],
        args_for_writing_out_config_file=['--dump'],
    )
    parser.add_argument("--test_dataset", default=None, help='path of the dataset we are going to run inference on')
    parser.add_argument("--pred_dataset", default=None, help='path of the dataset that we are going to write with predictions')
    parser.add_argument("--default_config_key", default=None, help='can be main or iarpa')
    parser.add_argument("--feat_dpath", type=str, help='path to dump asset files. If unspecified, choose a path adjacent to pred_dataset')
    # parser.add_argument("--tag", default='change_prob')
    # parser.add_argument("--package_fpath", type=pathlib.Path)
    parser.add_argument("--export_raw_features", default=smartcast, type=int, help='exporting raw features before classification head')

    # TODO: use torch packages instead
    parser.add_argument("--checkpoint_fpath", type=str, help='path to checkpoint file')
    parser.add_argument("--devices", default=None, help="lightning devices")

    parser.add_argument("--batch_size", default=1, type=int, help="prediction batch size")
    parser.add_argument("--num_workers", default=0, type=str, help="data loader workers, can be set to auto")

    parser.add_argument("--compress", default='DEFLATE', type=str, help="type of gdal compression to use")
    parser.add_argument("--blocksize", default=128, type=str, help="gdal COG blocksize")

    parser.add_argument("--cache", default=False, type=smartcast, help="if True enables caching of predictions in case of a crash")
    # parser.add_argument("--thresh", type=float, default=0.01)

    parser.set_defaults(**kwargs)
    default_args = None if cmdline else []
    args, _ = parser.parse_known_args(default_args)

    assert args.test_dataset is not None, 'must specify path to dataset to predict on'
    assert args.pred_dataset is not None, 'must specify path to dataset to predict on'
    assert args.checkpoint_fpath is not None, 'must specify the path to the checkpoint'

    return args


def hardcoded_default_configs(default_config_key):
    # HACK: THIS IS NOT ROBUST
    from watch.tasks import rutgers_material_seg
    from os.path import dirname, join
    module_dpath = dirname(rutgers_material_seg.__file__)
    main_config_path = join(module_dpath, "./configs/main.yaml")
    print('main_config_path = {!r}'.format(main_config_path))
    initial_config = utils.load_yaml_as_dict(main_config_path)
    experiment_config_path = join(module_dpath, f"./configs/{default_config_key}.yaml")
    experiment_config = utils.config_parser(experiment_config_path, experiment_type="training")
    config = {**initial_config, **experiment_config}
    config['start_time'] = datetime.datetime.today().strftime('%Y-%m-%d-%H:%M:%S')
    return config


def build_evaler(cmdline=False, **kwargs):
    """
    """
    args = make_predict_config(cmdline=cmdline, **kwargs)
    print('args.__dict__ = {}'.format(ub.urepr(args.__dict__, nl=1)))
    config = hardcoded_default_configs(args.default_config_key)

    config['start_time'] = datetime.datetime.today().strftime('%Y-%m-%d-%H:%M:%S')

    # Hacks to modify the config
    config['training']['pretrained'] = False
    print('config = {}'.format(ub.urepr(config, nl=2)))
    if 0:
        torch.manual_seed(config['seed'])
        torch.cuda.manual_seed(config['seed'])
        np.random.seed(config['seed'])
        random.seed(config['seed'])
        torch.set_default_dtype(torch.float32)

    from watch.utils.lightning_ext import util_device
    from watch.utils import util_resources
    devices = util_device.coerce_devices(args.devices)
    num_workers = util_parallel.coerce_num_workers(args.num_workers)
    if len(devices) > 1:
        print('args.devices = {!r}'.format(args.devices))
        print('devices = {!r}'.format(devices))
        raise NotImplementedError('TODO: handle multiple devices')
    device = devices[0]
    if num_workers > 0:
        util_resources.request_nofile_limits()
    print('device = {!r}'.format(device))
    print('num_workers = {!r}'.format(num_workers))

    # Load input kwcoco dataset and prepare the sampler
    input_coco_dset: kwcoco.CocoDataset = kwcoco.CocoDataset.coerce(args.test_dataset)

    # HACK: Filter to remove WV images
    invalid_sensors = {'WV'}
    orig_images = input_coco_dset.images()
    flags = [s not in invalid_sensors for s in orig_images.lookup('sensor_coarse')]
    valid_images = orig_images.compress(flags)
    input_coco_dset = input_coco_dset.subset(valid_images)

    channels = config['data']['channels']
    num_channels = len(channels.split('|'))
    config['training']['num_channels'] = num_channels

    # HACK!!!!
    # THIS IS WHY WE SAVE METADATA WITH THE MODEL!
    # WE DONT WANT TO HAVE TO FUDGE RECONSTRUCTION IN PRODUCTION!!!
    args.checkpoint_fpath = os.fspath(args.checkpoint_fpath)

    base_path = '/'.join(str(args.checkpoint_fpath).split('/')[:-1])
    pretrain_config_path = f"{base_path}/config.yaml"
    if os.path.isfile(pretrain_config_path):
        pretrain_config = utils.load_yaml_as_dict(pretrain_config_path)
        config['data']['channels'] = pretrain_config['data']['channels']
        # config['training']['model_feats_channels'] = pretrain_config_path['training']['model_feats_channels']

    # Load the model
    print('build model')
    model = build_model(model_name=config['training']['model_name'],
                        backbone=config['training']['backbone'],
                        pretrained=config['training']['pretrained'],
                        num_classes=config['data']['num_classes'],
                        num_groups=config['training']['gn_n_groups'],
                        weight_std=config['training']['weight_std'],
                        beta=config['training']['beta'],
                        num_channels=config['training']['num_channels'],
                        out_dim=config['training']['out_features_dim'],
                        feats=config['training']['model_feats_channels'])

    num_params = sum(p.numel() for p in model.parameters() if p.requires_grad)

    import netharn as nh
    mounted_model_cls = nh.device.DataSerial
    # mounted_model_cls = nn.DataParallel
    print("model has {} trainable parameters".format(num_params))
    model = mounted_model_cls(model)

    output_coco_fpath = pathlib.Path(args.pred_dataset)

    if args.feat_dpath is None:
        output_feat_dpath = output_coco_fpath.parent / '_assets/rutgers_material_seg'
    else:
        output_feat_dpath = pathlib.Path(args.feat_dpath)

    output_feat_dpath.mkdir(exist_ok=1, parents=True)
    output_coco_fpath.parent.mkdir(exist_ok=1, parents=True)

    print('init output dataset')
    # Create the results dataset as a copy of the test CocoDataset
    output_coco_dataset: kwcoco.CocoDataset = input_coco_dset.copy()
    print('clear output annots')
    # Remove all annotations in the results copy
    output_coco_dataset.clear_annotations()
    print('reroot output dset')
    # Change all paths to be absolute paths
    output_coco_dataset.reroot(absolute=True, check=False)
    output_coco_dataset.fpath = os.fspath(output_coco_fpath)

    imwrite_kw = {
        'compress': args.compress,
        'blocksize': args.blocksize,
    }

    print('make evaluator')
    evaler = Evaluator(
        model,
        checkpoint_fpath=args.checkpoint_fpath,
        input_coco_dset=input_coco_dset,
        output_coco_dataset=output_coco_dataset,
        config=config,
        device=device,
        num_workers=num_workers,
        output_feat_dpath=output_feat_dpath,
        imwrite_kw=imwrite_kw,
        batch_size=args.batch_size,
        cache=args.cache,
        save_raw_features=args.export_raw_features
        )
    return evaler


def main(cmdline=False, **kwargs):
    evaler = build_evaler(cmdline=cmdline, **kwargs)
    evaler.forward()


if __name__ == "__main__":
    main(cmdline=True)



# flake8: noqa

"""ResNet in PyTorch.
ImageNet-Style ResNet
[1] Kaiming He, Xiangyu Zhang, Shaoqing Ren, Jian Sun
    Deep Residual Learning for Image Recognition. arXiv:1512.03385
Adapted from: https://github.com/bearpaw/pytorch-classification
"""
import torch
import torch.nn as nn
import torch.nn.functional as F
from watch.tasks.rutgers_material_seg.models.encoding import Encoding
from watch.tasks.rutgers_material_seg.models.quantizer import Quantizer
from watch.tasks.rutgers_material_seg.models.tex_refine import TeRN
from torchvision import transforms


class ASPP(nn.Module):

    def __init__(self, C, depth, num_classes, conv=nn.Conv2d,
                 norm=nn.BatchNorm2d, momentum=0.0003, mult=1):
        super(ASPP, self).__init__()
        self._C = C
        self._depth = depth
        self._num_classes = num_classes

        self.global_pooling = nn.AdaptiveAvgPool2d(1)
        self.relu = nn.ReLU(inplace=True)
        self.aspp1 = conv(C, depth, kernel_size=1, stride=1, bias=False)
        self.aspp2 = conv(C, depth, kernel_size=3, stride=1,
                          dilation=int(6 * mult), padding=int(6 * mult),
                          bias=False)
        self.aspp3 = conv(C, depth, kernel_size=3, stride=1,
                          dilation=int(12 * mult), padding=int(12 * mult),
                          bias=False)
        self.aspp4 = conv(C, depth, kernel_size=3, stride=1,
                          dilation=int(18 * mult), padding=int(18 * mult),
                          bias=False)
        self.aspp5 = conv(C, depth, kernel_size=1, stride=1, bias=False)
        self.aspp1_bn = norm(depth, momentum)
        self.aspp2_bn = norm(depth, momentum)
        self.aspp3_bn = norm(depth, momentum)
        self.aspp4_bn = norm(depth, momentum)
        self.aspp5_bn = norm(depth, momentum)
        self.conv2 = conv(depth * 5, depth, kernel_size=1, stride=1,
                          bias=False)
        self.bn2 = norm(depth, momentum)
        self.conv3 = nn.Conv2d(depth, num_classes, kernel_size=1, stride=1)

    def forward(self, x):
        x1 = self.aspp1(x)
        x1 = self.aspp1_bn(x1)
        x1 = self.relu(x1)
        x2 = self.aspp2(x)
        x2 = self.aspp2_bn(x2)
        x2 = self.relu(x2)
        x3 = self.aspp3(x)
        x3 = self.aspp3_bn(x3)
        x3 = self.relu(x3)
        x4 = self.aspp4(x)
        x4 = self.aspp4_bn(x4)
        x4 = self.relu(x4)
        x5 = self.global_pooling(x)
        x5 = self.aspp5(x5)
        x5 = self.aspp5_bn(x5)
        x5 = self.relu(x5)
        x5 = nn.Upsample((x.shape[2], x.shape[3]), mode='bilinear',
                         align_corners=True)(x5)
        x = torch.cat((x1, x2, x3, x4, x5), 1)
        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu(x)
        x = self.conv3(x)

        return x


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1, is_last=False):
        super(BasicBlock, self).__init__()
        self.is_last = is_last
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion * planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion * planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        preact = out
        out = F.relu(out)
        if self.is_last:
            return out, preact
        else:
            return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, in_planes, planes, stride=1, is_last=False):
        super(Bottleneck, self).__init__()
        self.is_last = is_last
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, self.expansion * planes, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(self.expansion * planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion * planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion * planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(x)
        preact = out
        out = F.relu(out)
        if self.is_last:
            return out, preact
        else:
            return out


class ResNet(nn.Module):
    def __init__(self, block, num_blocks, num_channels=3, zero_init_residual=False,
                pretrained=False, num_classes=None, beta=False, weight_std=False,
                num_groups=32, out_dim=128, feats=[64, 64, 128, 256, 512]):
        super(ResNet, self).__init__()
        self.in_planes = 64
        self.num_codewords = 64

        self.pre_conv1 = nn.Conv2d(num_channels, 64, kernel_size=3, stride=1, padding=1,
                               bias=False)
        self.pre_bn1 = nn.BatchNorm2d(64)
        self.conv1 = nn.Conv2d(64 + num_channels, 64, kernel_size=3, stride=1, padding=1,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.layer1 = self._make_layer(block, 64, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, 128, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, 256, num_blocks[2], stride=2)
        self.layer4 = self._make_layer(block, 512, num_blocks[3], stride=2)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        # self.fc = nn.Linear(512 * block.expansion, num_classes)
        # self.aspp = ASPP(512, 256, 256)
        self.aspp = ASPP(2048, 256, 256)
        # self.quantizer = Quantizer(n_clusters=self.num_codewords, mode='euclidean', verbose=0, minibatch=None)

        self.encoding = nn.Sequential(
            Encoding(channels=256, num_codes=self.num_codewords),
            # nn.BatchNorm2d(self.num_codewords),
            nn.LeakyReLU(-0.2))
        # self.fc = nn.Sequential(nn.Linear(256, 256), nn.Sigmoid())

        self._aff = TeRN(num_iter=10, dilations=[1, 1, 2, 4, 6, 8])

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves
        # like an identity. This improves the model by 0.2~0.3% according to:
        # https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for i in range(num_blocks):
            stride = strides[i]
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def uncrop(self, cropped_image, params, H, W):
        bs, n_crops, c, h, w = cropped_image.shape
        uncrop = torch.zeros((bs, c, H, W), device=torch.device('cuda'))
        # uncrop_test = cropped_image.clone()
        # uncrop_test = uncrop_test.expand(-1,10,-1,-1,-1)
        # uncrop_test = torch.cat([uncrop_test, torch.zeros((bs,H*W-n_crops,c,h,w), device=torch.device('cuda'))], dim=1)
        # print(uncrop_test.shape)
        # with torch.no_grad():
        for b in range(bs):
            for crop in range(n_crops):
                top, left, height, width = params[crop]
                right = left + width
                bottom = top + height
                f_top, f_left = 0, 0
                f_right = f_left + width
                f_bottom = f_top + height

                if left < 0:
                    f_left = left - W
                    left = 0
                    # continue
                if top < 0:
                    f_top = top - H
                    top = 0
                    # continue
                if right > W:
                    f_right = W - left
                    right = W
                    # continue
                if bottom > H:
                    f_bottom = H - top
                    bottom = H
                    # continue

                # print(f"height: {height}, width:{width}")
                # print(f"left: {left}, top:{top}, right: {right}, bottom: {bottom} ")
                # print(f"f_left: {f_left}, f_top:{f_top}, f_right: {f_right}, f_bottom: {f_bottom} ")
                # features = cropped_image[b,crop,:,f_top:f_bottom, f_left:f_right]
                # print(features.shape)
                uncrop[b, :, top:bottom, left:right] += cropped_image[b, crop, :, f_top:f_bottom, f_left:f_right]
        # print(f"uncrop: {uncrop.dtype}")
        return uncrop

    def forward(self, x, original_image, sampled_crops):
        N, C, H, W = x.shape
        bs, c, h, w = original_image.shape

        # cropped_image = torch.stack([transforms.functional.crop(original_image, *params) for params in sampled_crops],dim=1)
        # bs, ps, pc, ph, pw = cropped_image.shape
        # cropped_image = cropped_image.view(bs*ps,pc,ph,pw)
        # x = x.view(bs*ps,c,ph,pw)
        # print(f"out: {out.dtype}")
        # print(out.requires_grad)
        # print(f"original_image: {original_image.shape}")
        out = F.relu(self.pre_bn1(self.pre_conv1(original_image)))

        refined_out = self._aff(original_image, out)
        # import matplotlib.pyplot as plt
        # import numpy as np
        # # out1_show = uncropped_out.sum(dim=1).cpu().detach().numpy()[0,:,:]
        # out1_show = refined_out.sum(dim=1).cpu().detach().numpy()[0,:,:]
        # # out1_show = uncropped_out.sum(dim=1).cpu().detach().numpy()[0,:,:]
        # out2_show = out.sum(dim=1).cpu().detach().numpy()[0,:,:]
        # x_show = np.transpose(original_image.cpu().detach().numpy()[0,:3,:,:], (1,2,0))
        # x_show = (x_show - x_show.min()) / (x_show.max() - x_show.min())
        # cmap_gradients = plt.cm.get_cmap('jet')
        # fig = plt.figure()
        # ax1 = fig.add_subplot(1,3,1)
        # ax2 = fig.add_subplot(1,3,2)
        # ax3 = fig.add_subplot(1,3,3)

        # ax1.imshow(x_show)
        # ax2.imshow(out1_show, cmap=cmap_gradients)
        # ax3.imshow(out2_show, cmap=cmap_gradients)
        # ax1.axis('off')
        # ax2.axis('off')
        # ax3.axis('off')
        # # plt.show()
        # image_name = np.random.randint(100000000000, size=1)[0]
        # plots_path_save = f"/home/native/projects/data/smart_watch/visualization/TextureRefinementNet"
        # fig_save_image_root = (f"{plots_path_save}/image_root/", ax1)
        # fig_save_out_root = (f"{plots_path_save}/out/", ax2)
        # fig_save_out_refined_root = (f"{plots_path_save}/out_refined/", ax3)

        # roots = [
        #     fig_save_image_root,
        #     fig_save_out_root,
        #     fig_save_out_refined_root,
        # ]

        # fig.savefig(
        #     f"{plots_path_save}/figs/{image_name}.png", bbox_inches='tight')
        # for root, ax in roots:
        #     file_path = f"{root}/{image_name}.png"
        #     # extent = ax.get_window_extent().transformed(figure.dpi_scale_trans.inverted())
        #     extent = ax.get_tightbbox(fig.canvas.get_renderer()).transformed(fig.dpi_scale_trans.inverted())
        #     fig.savefig(file_path, bbox_inches=extent)

        # fig.clear()
        # plt.cla()
        # plt.clf()
        # plt.close('all')
        # plt.close(fig)

        out = torch.stack([transforms.functional.crop(refined_out, *params) for params in sampled_crops], dim=1)
        bs, ps, pc, ph, pw = out.shape
        out = out.view(bs * ps, pc, ph, pw)

        # print(f"out: {out.shape}")
        # print(f"x: {x.shape}")
        out = torch.cat([out, x], dim=1)
        # print(f"out: {out.shape}")
        out = F.relu(self.bn1(self.conv1(out)))
        # print(f"out: {out.shape}")

        # out_clone = out.clone().detach()
        # print(f"out leaf: {out.is_leaf}")
        # with torch.no_grad():
        # with torch.no_grad():
        # chuncked_out = torch.stack(torch.chunk(out_clone, chunks=bs, dim=0), dim=0)
        # uncropped_out = self.uncrop(chuncked_out, params=sampled_crops, H=h, W=w)
        # # print(f"uncropped_out: {uncropped_out.shape}")
        # cropped_out = torch.stack([transforms.functional.crop(uncropped_out, *params) for params in sampled_crops], dim=1)
        # # print(f"cropped_out: {cropped_out.shape}")
        # bs, ps, pc, ph, pw = cropped_out.shape
        # cropped_out = cropped_out.view(bs*ps,pc,ph,pw)

        # print(cropped_out.requires_grad)
            # print(f"out: {out.shape}")
        # cropped_out.grad = out.grad
        # print(out.requires_grad)

        # print(f"out: {out.shape}")
        # print(f"uncropped_out: {uncropped_out.shape}")
        # print(f"uncropped_out: {uncropped_out.shape}")
        # print(f"cropped_image: {cropped_image.shape}")

        # print(f"out: {out.shape}")

        # out[:,:,:,:] = cropped_out.data

        # print(f"out: {out.shape}")
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = self.avgpool(out)
        out = self.aspp(out)
        # print(f"out: {out.shape}")
        out_enc = self.encoding(out)
        out_enc = out_enc.mean(dim=1)
        out = torch.flatten(out_enc, 1)
        # print(f"encoding_feat: {encoding_feat.shape}")
        # gamma = self.fc(encoding_feat)
        # print(f"gamma: {gamma.shape}")
        # y = gamma.unsqueeze().unsqueeze()
        # out = F.relu_(out + out * gamma)
        # out = torch.flatten(out, 1)
        # print(f"output: {output.shape}")

        # dictionary = self.quantizer.fit_predict(out)
        # centroids = self.quantizer.centroids.T
        # print(f"centroids: {centroids.shape}")
        # residuals = torch.cdist(out, centroids.T)
        # print(f"residuals: {residuals.shape}")
        # quant =  torch.tensor([torch.histc(residuals[i], bins=self.num_codewords).tolist() for i in range(N)], device=torch.device('cuda'), requires_grad=True)

        return out  # , out

    # def forward(self, x, layer=100):
    #     N, C, H, W = x.shape
    #     recon_img = torch.stack(torch.chunk(x, chunks=4, dim=0), dim=0).view(4,1024,9,-1)#.mean(dim=-1)
    #     print(recon_img.shape)
    #     print(f"x: {x.shape}")
    #     out = F.relu(self.bn1(self.conv1(x)))
    #     print(f"out: {out.shape}")
    #     out = self._aff(x, out)
    #     print(f"out: {out.shape}")
    #     out = self.layer1(out)
    #     out = self.layer2(out)
    #     out = self.layer3(out)
    #     out = self.layer4(out)
    #     out = self.avgpool(out)
    #     out = self.aspp(out)
    #     # print(f"out: {out.shape}")
    #     encoding_feat = self.encoding(out).mean(dim=1)
    #     out = torch.flatten(out, 1)
    #     # print(f"encoding_feat: {encoding_feat.shape}")
    #     # gamma = self.fc(encoding_feat)
    #     # print(f"gamma: {gamma.shape}")
    #     # y = gamma.unsqueeze().unsqueeze()
    #     # out = F.relu_(out + out * gamma)
    #     # out = torch.flatten(out, 1)
    #     # print(f"output: {output.shape}")

    #     # dictionary = self.quantizer.fit_predict(out)
    #     # centroids = self.quantizer.centroids.T
    #     # print(f"centroids: {centroids.shape}")
    #     # residuals = torch.cdist(out, centroids.T)
    #     # print(f"residuals: {residuals.shape}")
    #     # quant =  torch.tensor([torch.histc(residuals[i], bins=self.num_codewords).tolist() for i in range(N)], device=torch.device('cuda'), requires_grad=True)

    #     return out, encoding_feat


def resnet18(**kwargs):
    return ResNet(BasicBlock, [2, 2, 2, 2], **kwargs)


def resnet34(**kwargs):
    return ResNet(BasicBlock, [3, 4, 6, 3], **kwargs)


def resnet50(**kwargs):
    return ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)


def resnet101(**kwargs):
    return ResNet(Bottleneck, [3, 4, 23, 3], **kwargs)


model_dict = {
    'resnet18': [resnet18, 512],
    'resnet34': [resnet34, 512],
    'resnet50': [resnet50, 2048],
    'resnet101': [resnet101, 2048],
}


class LinearBatchNorm(nn.Module):
    """Implements BatchNorm1d by BatchNorm2d, for SyncBN purpose"""

    def __init__(self, dim, affine=True):
        super(LinearBatchNorm, self).__init__()
        self.dim = dim
        self.bn = nn.BatchNorm2d(dim, affine=affine)

    def forward(self, x):
        x = x.view(-1, self.dim, 1, 1)
        x = self.bn(x)
        x = x.view(-1, self.dim)
        return x


class SupConResNet(nn.Module):
    """backbone + projection head"""

    def __init__(self, name='resnet50', head='mlp', feat_dim=128):
        super(SupConResNet, self).__init__()
        model_fun, dim_in = model_dict[name]
        self.encoder = model_fun()
        if head == 'linear':
            self.head = nn.Linear(dim_in, feat_dim)
        elif head == 'mlp':
            self.head = nn.Sequential(
                nn.Linear(dim_in, dim_in),
                nn.ReLU(inplace=True),
                nn.Linear(dim_in, feat_dim)
            )
        else:
            raise NotImplementedError(
                'head not supported: {}'.format(head))

    def forward(self, x):
        feat = self.encoder(x)
        feat = F.normalize(self.head(feat), dim=1)
        return feat


class SupCEResNet(nn.Module):
    """encoder + classifier"""

    def __init__(self, name='resnet50', num_classes=10):
        super(SupCEResNet, self).__init__()
        model_fun, dim_in = model_dict[name]
        self.encoder = model_fun()
        self.fc = nn.Linear(dim_in, num_classes)

    def forward(self, x):
        return self.fc(self.encoder(x))


class LinearClassifier(nn.Module):
    """Linear classifier"""

    def __init__(self, name='resnet50', num_classes=10):
        super(LinearClassifier, self).__init__()
        _, feat_dim = model_dict[name]
        self.fc = nn.Linear(feat_dim, num_classes)

    def forward(self, features):
        return self.fc(features)



import os
import pytorch_lightning as pl
from argparse import ArgumentParser, Namespace
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.callbacks import ModelCheckpoint, LearningRateMonitor
from .utils import setup_python_logging
from .time_sort_module import time_sort


def main(args):
    if isinstance(args, dict):
        args = Namespace(**args)
    log_dir = '{}/{}/{}/train_video_{}'.format(
        args.save_dir,
        'temporal_sequence_predict',
        args.sensor,
        args.train_video
    )
    exp_name = 'default'
    logger = TensorBoardLogger(log_dir, name=exp_name)

    setup_python_logging(logger.log_dir)

    model = time_sort(hparams=args)

    checkpoint_callback = ModelCheckpoint(
        monitor='val_acc', mode='max', save_top_k=2)
    lr_logger = LearningRateMonitor(logging_interval='step')

    trainer = pl.Trainer.from_argparse_args(
        args, logger=logger, callbacks=[
            checkpoint_callback, lr_logger])
    trainer.fit(model)


if __name__ == '__main__':
    parser = ArgumentParser()

    parser.add_argument('--max_epochs', type=int, default=100)
    parser.add_argument('--workers', type=int, default=8)
    parser.add_argument('--batch_size', type=int, default=16)
    parser.add_argument('--step_size', type=int, default=20)
    parser.add_argument('--learning_rate', type=float, default=.0003)
    parser.add_argument('--gamma', type=float, default=.1)
    parser.add_argument('--weight_decay', type=float, default=1e-5)
    parser.add_argument('--save_dir', type=str, default='logs')
    parser.add_argument('--gpus', type=int)
    parser.add_argument('--feature_dim', type=int, default=64)
    parser.add_argument(
        '--backbone',
        help='choose from unet, unet_blur',
        default='unet_blur')

    parser.add_argument(
        '--panchromatic',
        help='set flag for using panchromatic WV imagery',
        action='store_true')
    parser.add_argument(
        '--sensor',
        type=str,
        help='choose from WV, LC, or S2',
        default='S2')

    parser.add_argument(
        '--in_channels',
        help='specify the number of channels corresponding to the sensor type',
        type=int,
        default=3)
    parser.add_argument('--train_video', type=int, default=3)
    parser.add_argument('--val_video', type=int, default=5)
    parser.add_argument(
        '--min_time_step',
        help='enforce minimum distance between image pairs',
        type=int,
        default=1)

    parser.add_argument(
        '--train_dataset',
        type=str,
        default='/u/eag-d1/data/watch/drop0_aligned/data.kwcoco.json')

    parser.add_argument(
        '--val_dataset',
        type=str,
        default='/u/eag-d1/data/watch/drop0_aligned/data.kwcoco.json')

    parser.set_defaults(
        gpus=1,
        terminate_on_nan=True,
        check_val_every_n_epochs=1,
        log_every_n_steps=20,
        flush_logs_every_n_steps=20,
        panchromatic=False
    )

    args = parser.parse_args()
    args.default_save_path = os.path.join(args.save_dir, "logs")

    main(args)



import pandas as pd
import numpy as np
import ubelt as ub
import kwplot
import kwimage
from matplotlib.collections import LineCollection
from matplotlib.colors import to_rgba
from matplotlib.dates import date2num
from typing import Literal
import datetime as datetime_mod


def viz_sc(region_dpaths, true_site_dpath, true_region_dpath, save_dpath):
    """
    Seems broken in 1.0.0

    Ignore:
        >>> # xdoctest: +REQUIRES(module:iarpa_smart_metrics)
        >>> from watch.cli.run_metrics_framework import *  # NOQA
        >>> from watch.demo.metrics_demo.generate_demodata import generate_demo_metrics_framework_data
        >>> cmdline = 0
        >>> base_dpath = ub.Path.appdir('watch', 'tests', 'test-iarpa-metrics2')
        >>> data_dpath = base_dpath / 'inputs'
        >>> dpath = base_dpath / 'outputs'
        >>> demo_info3 = generate_demo_metrics_framework_data(
        >>>     roi='DR_R003',
        >>>     num_sites=11, num_observations=10, noise=3, p_observe=0.5,
        >>>     p_transition=0.2, drop_noise=0.3, drop_limit=0.5)
        >>> print('demo_info3 = {}'.format(ub.urepr(demo_info3, nl=1)))
        >>> out_dpath = dpath / 'region_metrics'
        >>> merge_fpath = dpath / 'merged.json'
        >>> out_dpath.delete()
        >>> kwargs = {
        >>>     'pred_sites': demo_info3['pred_site_dpath'],
        >>>     'true_region_dpath': demo_info3['true_region_dpath'],
        >>>     'true_site_dpath': demo_info3['true_site_dpath'],
        >>>     'merge': True,
        >>>     'merge_fpath': merge_fpath,
        >>>     'out_dir': out_dpath,
        >>> }
        >>> main(cmdline=False, **kwargs)
        >>> #
        >>> from watch.tasks.metrics.viz_sc_results import *  # NOQA
        >>> region_dpaths = [save_dpath / 'region_metrics/DR_R003']
        >>> save_dpath = merge_fpath.parent
        >>> true_site_dpath = demo_info3['true_site_dpath']
        >>> true_region_dpath = demo_info3['true_region_dpath']
        >>> viz_sc(region_dpaths, true_site_dpath, true_region_dpath, save_dpath)
        >>> # TODO: visualize
    """
    from watch.tasks.metrics.merge_iarpa_metrics import RegionResult
    kwplot.autompl()

    results = []
    for pth in region_dpaths:
        try:
            results.append(
                RegionResult.from_dpath_and_anns_root(
                    pth, true_site_dpath, true_region_dpath)
            )
        except FileNotFoundError:
            print(f'warning: missing region {pth}')

    # merge SC
    sc_results = [r for r in results if r.sc_dpath]

    # check out:
    # kwimage.stack_image
    # kwplot.make_legend_image

    phs = [ph for r in sc_results if (ph := r.sc_phasetable) is not None]

    for ph in ub.ProgIter(phs, desc='visualize sc gantt'):

        rid = ph.index.get_level_values('region_id')[0]

        # site-level viz
        for (site, site_cand), df_ in ph.groupby(['site', 'site_candidate']):
            df = df_.droplevel([0, 1, 2])
            viz_sc_gantt(
                df,
                ' vs. '.join((site, site_cand)),
                ((save_dpath / rid).ensuredir() / ('_'.join((site, site_cand)) + '.png'))
            )

        # region-level viz
        df = ph
        plot_title = rid
        save_fpath = (save_dpath / f'sc_{rid}.png')
        viz_sc_multi(ph, plot_title, save_fpath)

    # merged viz
    print('Visualize merged')
    merged_df = pd.concat(phs, axis=0)
    viz_sc_multi(
        merged_df,
        ' '.join(merged_df.index.unique(level='region_id')),
        (save_dpath / 'sc_merged.png')
    )


def viz_sc_gantt(df, plot_title, save_fpath):
    from watch.heuristics import PHASES as phases
    plt = kwplot.autoplt()
    sns = kwplot.autosns()
    # TODO how to pick site boundary?
    df = df.apply(lambda s: s.str.split(', '))

    df['pred'] = df['pred'].fillna(method='ffill')
    # df = df[~df['true'].isna()]
    # df['pred'] = df['pred'].fillna('Unknown')
    df['pred'] = df['pred'].fillna(method='bfill')
    df = df.reset_index()
    df['date'] = pd.to_datetime(df['date']).dt.date
    df = df.melt(id_vars=['date'])
    df = df.dropna()

    # split out subsites
    subsite_ixs = df['value'].str.len() > 1
    var, val = [], []
    for _, row in df.loc[subsite_ixs].iterrows():
        for v in row['value']:
            var.append(row['variable'])
            val.append(v)
    df = df.loc[~subsite_ixs]
    df['value'] = df['value'].str[0]
    df = pd.concat(
        (df,
         pd.DataFrame(dict(variable=var, value=val))),
        axis=0,
        ignore_index=True,
    )

    # order hack for relplot
    phases_type = pd.api.types.CategoricalDtype(
        (['Unknown'] + phases)[::-1], ordered=True)
    df['value'] = df['value'].astype(phases_type)
    df = df.sort_values(by='value')

    # TODO threadsafe
    grid = sns.relplot(
        data=df,
        x='date',
        y='value',
        hue='variable',
        size='variable'
    )
    plt.title(plot_title)
    grid.savefig(save_fpath)
    plt.close()


def viz_sc_multi(ph, plot_title, save_fpath,
                 date: Literal['absolute', 'from_start', 'from_active'] = 'absolute',
                 how: Literal['residual', 'strip'] = 'strip'):
    from watch.heuristics import PHASES as phases
    import watch.heuristics
    plt = kwplot.autoplt()
    sns = kwplot.autosns()
    df = ph

    # df.index = [df.index.map('{0[0]} {0[1]} {0[2]}'.format), df.index.get_level_values(3)]

    # ignore subsites
    # TODO how to pick site boundary?
    df = df.apply(lambda s: s.str.split(', ').str[0])

    # could do just region_id for error bars after fillna
    df = df.reset_index()
    df['group'] = df[['region_id', 'site', 'site_candidate']].astype('string').agg('\n'.join, axis=1)
    df = df.drop(['region_id', 'site', 'site_candidate'], axis=1)

    if how == 'residual':  # true and pred must be aligned
        df['pred'] = (df.groupby('group')['pred']
                        .fillna(method='ffill')
                        .fillna('No Activity'))
        # df['pred'] = df['pred'].fillna('Unknown')
        # df['pred'] = df['pred'].fillna(method='bfill')

        df = df[~df['true'].isna()]
        # df = df[df['true'] != 'Unknown']  # Unk should be gone from df after this

    df['date'] = pd.to_datetime(df['date'])  # .dt.date

    # must do this before searchsorted
    phases_type = pd.api.types.CategoricalDtype(
        ['Unknown'] + phases, ordered=True)
    df['pred'] = df['pred'].astype(phases_type)
    df['true'] = df['true'].astype(phases_type)
    # assert df.groupby('group')['true'].is_monotonic_increasing.all()
    # assert df.groupby('group')['pred'].is_monotonic_increasing.all()

    if date == 'absolute':  # absolute date
        df['date'] = df['date'].dt.date
        x_var = 'date'
    elif date == 'from_start':  # relative date since start
        df['date'] = df.groupby('group')['date'].transform(lambda date: date - date.iloc[0])
        # df['date'] = df['date'].dt.days.fillna(0)
        df['date'] = df['date'].dt.days
        x_var = 'days since start'
    elif date == 'from_active':  # relative date since last NA; requires 1 NA to exist before SP
        df = df.groupby('group').apply(_align_start)
        df['date'] = df['date'].dt.days
        x_var = 'days since final No Activity'

    palette = {c['name']: c['color'] for c in watch.heuristics.CATEGORIES}

    if how == 'residual':
        jitter = 0.1
        df['diff'] = df['pred'].cat.codes - df['true'].cat.codes
        df['diff'] += np.random.uniform(low=-jitter, high=jitter, size=len(df['diff']))

        # TODO threadsafe
        grid = sns.relplot(
            kind='scatter',
            data=df,
            x='date',
            y='diff',
            hue='true',
            palette=palette,
            s=10,
        )
        grid.map(_add_colored_linesegments,
                 'date', 'diff', 'true', 'group',)
        y_var = 'pred phases ahead of true phase'
    elif how == 'strip':
        # get tp idxs before reshaping
        tps = df.groupby('group').apply(_tp_idxs)

        dfm = df.melt(id_vars=['date', 'group'], value_name='phase').dropna()
        dfm['phase'] = dfm['phase'].astype(phases_type)
        dfm['yval'], ylabels = pd.factorize(dfm['group'])
        tps.index = tps.index.map(dict(zip(ylabels, range(len(ylabels)))))
        with pd.option_context('mode.chained_assignment', None):
            dfm['yval'].loc[dfm['variable'] == 'pred'] -= 0.2
        grid = sns.relplot(
            kind='scatter',
            data=dfm,
            x='date',
            y='yval',
            hue='phase',
            palette=palette,
            s=10,
            facet_kws=dict(legend_out=False),
        )
        grid.map(_add_colored_linesegments,
                 'date', 'yval', 'phase', 'yval', palette=palette)

        grid.map(_highlight_tp, 'yval', tps=tps)
        if len(ylabels) <= 20:  # draw site names if they'll be readable
            grid.set(yticks=range(len(ylabels)))
            grid.set_yticklabels(ylabels, size=4)
        # and write them as an index
        yregion, ytrue, ypred = np.array(ylabels.str.split('\n').to_list()).T
        pd.DataFrame(dict(
            index=range(len(ylabels)),
            region=yregion,
            true=ytrue,
            pred=ypred,
        )).to_csv(save_fpath.with_suffix('.index.csv'))
        y_var = '[region, true, pred]'

    # df['group_phase'] = df[['group', 'true']].agg('_'.join, axis=1)

    sns.move_legend(grid, 'upper right')
    grid.set_axis_labels(x_var=x_var, y_var=y_var)
    plt.title(plot_title)
    grid.savefig(save_fpath)
    plt.close()


def _align_start(grp, phase='Site Preparation', before=True):
    grp = grp.sort_values(by='date')
    # assert grp['true'].iloc[0] != phase, grp['true']
    grp['date'] -= grp.iloc[grp['true'].searchsorted(phase) - int(before)]['date']
    return grp


def _tp_idxs(grp):
    grp = grp.sort_values(by='date')
    grp['pred'] = (grp['pred']
                   .fillna(method='ffill')
                   .fillna('No Activity'))
    grp = grp[~grp['true'].isna()]
    match = grp['pred'] == grp['true']
    ixs = grp[match.shift() != match].index.values
    x = grp['date'].map(date2num)
    blocks = [x.loc[[start, end]] for start, end, matches in zip(ixs, ixs[1:], match[ixs]) if matches]
    return blocks


def _highlight_tp(y, **kwargs):
    tps = kwargs.get('tps')
    sites = y.round().abs().astype(int).unique()
    for site in sites:
        boxes = kwimage.Boxes([[*xs, site - 0.5, site + 0.5] for xs in tps.loc[site]], format='xxyy')
        kwplot.draw_boxes(boxes, color='green', fill=True, lw=0, alpha=0.3)


def _add_colored_linesegments(x, y, phase, units, **kwargs):
    """
    x = dfm['date']
    y = dfm['yval']
    phase = dfm['phase']
    units = dfm['yval']
    """
    # need args instead of kwargs because of grid.map() weirdness
    plt = kwplot.autoplt()
    palette = kwargs.get('palette')

    if isinstance(x.iloc[0], datetime_mod.date):
        x = date2num(x)

    _df = pd.DataFrame(dict(
        xy=zip(x, y),
        hue=pd.Series(phase).map(palette).astype('string').map(to_rgba),
        phase=phase,  # need to keep this due to float comparisons in searchsorted
        units=units,
    ))

    lines = []
    colors = []
    for unit, grp in _df.groupby('units'):
        # drop consecutive dups
        ph = grp['phase']
        ixs = ph.loc[ph.shift() != ph].index.values
        for start, end, hue in zip(
                ixs,
                ixs[1:].tolist() + [None],
                grp['hue'].loc[ixs]
        ):
            line = grp['xy'].loc[start:end].values.tolist()
            if len(line) > 0:
                lines.append(line)
                colors.append(hue)

    lc = LineCollection(lines, alpha=0.5, colors=colors)
    ax = plt.gca()
    ax.add_collection(lc)
    # ax.autoscale()



def fix_torchmetrics_compatability():
    import torchmetrics
    from watch.monkey._monkey_fbeta import FBetaScore_Patched
    from watch.monkey._monkey_fbeta import Accuracy_Patched
    f_beta = torchmetrics.classification.f_beta
    accuracy = torchmetrics.classification.accuracy

    if not hasattr(f_beta, 'FBetaScoreOrig'):
        f_beta.FBetaScoreOrig = f_beta.FBetaScore

    if not hasattr(accuracy, 'FBetaScoreOrig'):
        accuracy.AccuracyOrig = accuracy.Accuracy

    f_beta.FBetaScore = FBetaScore_Patched
    f_beta.F1 = FBetaScore_Patched

    accuracy.Accuracy = Accuracy_Patched



import torch

from watch.tasks.rutgers_material_change_detection.models.base_model import BaseFramework


class PatchTransformerFramework(BaseFramework):
    def __init__(self, task_mode, encoder, decoder, **kwargs):
        super(PatchTransformerFramework, self).__init__(task_mode, encoder, decoder)

        self.first_last_feat_aux_loss = kwargs["first_last_feat_aux_loss"]

    def forward(self, data):
        video = data["video"]  # [batch_size, frames, channels, height, width]

        # Pass input into the encoder.
        feat_tokens = self.encoder(video)  # TODO: Add datetimes here.

        # Pass input into the decoder.
        if self.task_mode in ["total_bin_change", "total_sem_change"]:
            # Generate a prediction image from patches.
            height = feat_tokens[0]["og_height"]
            width = feat_tokens[0]["og_width"]
            batch_size, n_out_channels = self.decoder(feat_tokens[0]["prediction"]).shape[:2]
            device = self.decoder(feat_tokens[0]["prediction"])
            canvas = torch.zeros([batch_size, n_out_channels, height, width]).to(device)

            first_token_feats, last_token_feats = [], []
            for feat_token in feat_tokens:
                # Get prediciton.
                final_preds = self.decoder(feat_token["prediction"])

                # Fill prediction into output canvas.
                H0, W0, dH, dW = feat_token["crop_slice"]
                canvas[:, :, H0 : H0 + dH, W0 : W0 + dW] = final_preds

                if self.first_last_feat_aux_loss:
                    first_token_feats.append(feat_token["prediction"][:, 0].flatten(1))
                    last_token_feats.append(feat_token["prediction"][:, -2].flatten(1))

        elif self.task_mode in ["ss_splice_change", "ss_splice_change_index", "ss_arrow_of_time"]:
            # Generate a final prediction from all guesses.
            batch_size, n_out_channels = self.decoder(feat_tokens[0]["prediction"]).shape[:2]
            device = self.decoder(feat_tokens[0]["prediction"])
            canvas = torch.zeros([batch_size, n_out_channels]).to(device)

            first_token_feats, last_token_feats = [], []
            for feat_token in feat_tokens:
                canvas += feat_token["prediction"]

                if self.first_last_feat_aux_loss:
                    first_token_feats.append(feat_token["prediction"][:, 0].flatten(1))
                    last_token_feats.append(feat_token["prediction"][:, -2].flatten(1))
        else:
            raise NotImplementedError(
                f'Task mode "{self.task_mode}" not implemented for Patch Tranformer Framework forward method.'
            )

        output = {}
        output[self.task_mode] = canvas
        output["first_token_feat"] = first_token_feats
        output["first_token_feat"] = last_token_feats
        return output



import tempfile
import subprocess
import json
import os
from urllib.parse import urlparse

import pystac

from watch.cli.baseline_framework_ingress import ingress_item
from watch.cli.baseline_framework_egress import egress_item


class CacheItemOutputS3Wrapper:
    def __init__(self, item_map, outbucket, aws_profile=None):
        self.item_map = item_map
        self.outbucket = outbucket

        if aws_profile is not None:
            self.aws_base_command = [
                'aws', 's3', '--profile', aws_profile, 'cp', '--no-progress']
        else:
            self.aws_base_command = ['aws', 's3', 'cp', '--no-progress']

    def __call__(self, stac_item, *args, **kwargs):
        with tempfile.TemporaryDirectory() as tmpdirname:
            status_file_basename = '{}.done'.format(stac_item['id'])
            status_item_s3_path = os.path.join(
                self.outbucket, 'status', status_file_basename)
            status_item_local_path = os.path.join(
                tmpdirname, status_file_basename)

            try:
                subprocess.run([*self.aws_base_command,
                                status_item_s3_path,
                                status_item_local_path],
                               check=True)
            except subprocess.CalledProcessError:
                pass
            else:
                print("* Item: {} previously processed, not "
                      "re-processing".format(stac_item['id']))
                with open(status_item_local_path) as f:
                    return [json.loads(line) for line in f]

            output_stac_items = self.item_map(stac_item, *args, **kwargs)

            output_status_file = os.path.join(
                tmpdirname, '{}.output.done'.format(stac_item['id']))
            with open(output_status_file, 'w') as outf:
                if isinstance(output_stac_items, dict):
                    print(json.dumps(output_stac_items), file=outf)
                elif isinstance(output_stac_items, pystac.Item):
                    print(json.dumps(output_stac_items.to_dict()), file=outf)
                else:
                    for output_item in output_stac_items:
                        if isinstance(output_item, pystac.Item):
                            print(json.dumps(output_item.to_dict()), file=outf)
                        else:
                            print(json.dumps(output_item), file=outf)

            subprocess.run([*self.aws_base_command,
                            output_status_file,
                            status_item_s3_path], check=True)

            return output_stac_items


def _default_item_selector(stac_item):
    return True


def _default_asset_selector(asset_name, asset):
    return True


class IngressProcessEgressWrapper:
    def __init__(self,
                 item_map,
                 outbucket,
                 aws_base_command,
                 dryrun=False,
                 stac_item_selector=_default_item_selector,
                 asset_selector=_default_asset_selector,
                 skip_egress=False):
        self.item_map = item_map
        self.outbucket = outbucket
        self.aws_base_command = aws_base_command
        self.dryrun = dryrun
        self.stac_item_selector = stac_item_selector
        self.asset_selector = asset_selector
        self.skip_egress = skip_egress

    def __call__(self, stac_item, *args, **kwargs):
        # Assumes that the 'self.item_map' function accepts
        # 'stac_item' and 'working_dir' arguments. TODO: actually
        # check this via introspection
        print("* Processing item: {}".format(stac_item['id']))

        if not self.stac_item_selector(stac_item):
            print("**  STAC item {} did not satisfy selector, not "
                  "processing".format(stac_item['id']))
            return [stac_item]

        with tempfile.TemporaryDirectory() as tmpdirname:
            ingressed_item = ingress_item(
                stac_item,
                os.path.join(tmpdirname, 'ingress'),
                self.aws_base_command,
                self.dryrun,
                relative=False,
                asset_selector=self.asset_selector)

            # Stripping the 'root' link here as it usually refers to
            # the catalog which isn't ingressed when we call
            # ingress_item directly (can throw exceptions when trying
            # to convert to dict or serialize when the catalog is
            # missing)
            ingressed_item.remove_links('root')

            processed_item = self.item_map(
                ingressed_item,
                tmpdirname,
                *args, **kwargs)

            processed_items = []
            if isinstance(processed_item, dict):
                processed_items.append(pystac.Item.from_dict(processed_item))
            elif isinstance(processed_item, pystac.Item):
                processed_items.append(processed_item)
            else:
                # Assume already an iterable of pystac.Item
                processed_items = processed_item

            if self.skip_egress:
                return processed_items

            output_items = []
            for item in processed_items:
                output_items.append(egress_item(item,
                                                self.outbucket,
                                                self.aws_base_command))

            # Returning a list here
            return output_items


def download_region(input_region_path,
                    output_region_path,
                    aws_profile=None,
                    strip_nonregions=False,
                    ensure_comments=False):
    if aws_profile is not None:
        aws_base_command =\
            ['aws', 's3', '--profile', aws_profile, 'cp']
    else:
        aws_base_command = ['aws', 's3', 'cp']

    scheme, *_ = urlparse(input_region_path)
    if scheme == 's3':
        with tempfile.NamedTemporaryFile() as temporary_file:
            command = [*aws_base_command,
                       input_region_path,
                       temporary_file.name]

            print("Running: {}".format(' '.join(command)))
            # TODO: Manually check return code / output
            subprocess.run(command, check=True)

            with open(temporary_file.name) as f:
                out_region_data = json.load(f)
    elif scheme == '':
        with open(input_region_path) as f:
            out_region_data = json.load(f)
    else:
        raise NotImplementedError("Don't know how to pull down region file "
                                  "with URI scheme: '{}'".format(scheme))

    if strip_nonregions:
        out_region_data['features'] =\
            [feature
             for feature in out_region_data.get('features', ())
             if ('properties' in feature
                 and feature['properties'].get('type') == 'region')]

    if ensure_comments:
        # Ensure the region feature has a "comments" field
        for feature in out_region_data.get('features', ()):
            props = feature['properties']
            if props['type'] == 'region':
                props['comments'] = props.get('comments', '')

    with open(output_region_path, 'w') as f:
        print(json.dumps(out_region_data, indent=2), file=f)

    return output_region_path


def determine_region_id(region_fpath):
    """
    Args:
        region_fpath (str | PathLike):
            the path to a region model geojson file

    Returns:
        str | None : the region id if we can find one
    """
    region_id = None
    with open(region_fpath, 'r') as file:
        region_data = json.load(file)
        for feature in region_data.get('features', []):
            props = feature['properties']
            if props['type'] == 'region':
                region_id = props.get('region_id', props.get('region_model_id'))
                break
    return region_id



#!/usr/bin/env python3
import os
import subprocess
import json

from watch.cli.baseline_framework_kwcoco_ingress import baseline_framework_kwcoco_ingress  # noqa: 501
from watch.cli.baseline_framework_kwcoco_egress import baseline_framework_kwcoco_egress  # noqa: 501
from watch.utils.util_framework import download_region
import ubelt as ub
import scriptconfig as scfg
from watch.mlops.smart_pipeline import SV_Cropping
from watch.utils.util_yaml import Yaml


class SVDatasetConfig(scfg.DataConfig):
    """
    Generate cropped KWCOCO dataset for SC
    """
    input_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
            '''
            Path to input T&E Baseline Framework JSON
            '''))
    input_region_path = scfg.Value(None, type=str, position=2, required=True, help=ub.paragraph(
            '''
            Path to input T&E Baseline Framework Region definition JSON
            '''))
    output_path = scfg.Value(None, type=str, position=3, required=True, help='S3 path for output JSON')
    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            AWS Profile to use for AWS S3 CLI commands
            '''))
    dryrun = scfg.Value(False, isflag=True, short_alias=['d'], help='Run AWS CLI commands with --dryrun flag')
    outbucket = scfg.Value(None, type=str, required=True, short_alias=['o'], help=ub.paragraph(
            '''
            S3 Output directory for STAC item / asset egress
            '''))
    newline = scfg.Value(False, isflag=True, short_alias=['n'], help=ub.paragraph(
            '''
            Output as simple newline separated STAC items
            '''))
    jobs = scfg.Value(1, type=int, short_alias=['j'], help='Number of jobs to run in parallel')
    dont_recompute = scfg.Value(False, isflag=True, help=ub.paragraph(
            '''
            Will not recompute if output_path already exists
            '''))
    force_one_job_for_cropping = scfg.Value(False, isflag=True, help='Force jobs=1 for cropping')
    sv_cropping_config = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            Raw json/yaml or a path to a json/yaml file that specifies the
            config for SV_Cropping.
            '''))


def main():
    config = SVDatasetConfig.cli(strict=True)
    print('config = {}'.format(ub.urepr(dict(config), nl=1, align=':')))
    run_generate_sv_cropped_kwcoco(**config)


def run_generate_sv_cropped_kwcoco(input_path,
                                   input_region_path,
                                   output_path,
                                   outbucket,
                                   aws_profile=None,
                                   dryrun=False,
                                   newline=False,
                                   jobs=1,
                                   dont_recompute=False,
                                   force_one_job_for_cropping=False,
                                   sv_cropping_config=None):
    if dont_recompute:
        if aws_profile is not None:
            aws_ls_command = ['aws', 's3', '--profile', aws_profile, 'ls']
        else:
            aws_ls_command = ['aws', 's3', 'ls']

        try:
            subprocess.run([*aws_ls_command, output_path], check=True)
        except subprocess.CalledProcessError:
            # Continue processing
            pass
        else:
            # If output_path file was there, nothing to do
            return

    # 1. Ingress data
    print("* Running baseline framework kwcoco ingress *")
    ingress_dir = '/tmp/ingress'
    _ = baseline_framework_kwcoco_ingress(
        input_path,
        ingress_dir,
        aws_profile,
        dryrun)

    # 2. Download and prune region file
    print("* Downloading and pruning region file *")
    local_region_path = '/tmp/region.json'
    local_region_path = download_region(input_region_path,
                                        local_region_path,
                                        aws_profile=aws_profile,
                                        strip_nonregions=True)

    # Parse region_id from original region file
    with open(local_region_path) as f:
        region = json.load(f)

        region_id = None
        for feature in region.get('features', ()):
            props = feature['properties']
            if props['type'] == 'region':
                region_id = props.get('region_model_id',
                                      props.get('region_id'))
                break

    if region_id is None:
        raise RuntimeError("Couldn't parse 'region_id' from input region file")

    # Paths to inputs generated in previous pipeline steps
    bas_region_path = os.path.join(ingress_dir,
                                   'cropped_region_models_bas',
                                   '{}.geojson'.format(region_id))
    ta1_sc_kwcoco_path = os.path.join(ingress_dir,
                                      'kwcoco_for_sc.json')

    # 4. Crop ingress KWCOCO dataset to region for SV
    print("* Cropping KWCOCO dataset to region for SV*")
    ta1_sv_cropped_kwcoco_path = os.path.join(ingress_dir,
                                              'cropped_kwcoco_for_sv.json')

    sv_cropping_config = Yaml.coerce(sv_cropping_config or {})

    sv_cropping = SV_Cropping(root_dpath='/tmp/ingress')
    sv_cropping.configure({
        'crop_src_fpath': ta1_sc_kwcoco_path,
        'regions': bas_region_path,
        'crop_dst_fpath': ta1_sv_cropped_kwcoco_path})

    ub.cmd(sv_cropping.command(), check=True, verbose=3, system=True)

    # 5. Egress (envelop KWCOCO dataset in a STAC item and egress;
    #    will need to recursive copy the kwcoco output directory up to
    #    S3 bucket)
    print("* Egressing KWCOCO dataset and associated STAC item *")
    baseline_framework_kwcoco_egress(ta1_sv_cropped_kwcoco_path,
                                     local_region_path,
                                     output_path,
                                     outbucket,
                                     aws_profile=None,
                                     dryrun=False,
                                     newline=False)


if __name__ == "__main__":
    main()



import torch
import torch.nn as nn
import torch.nn.functional as F


class Pre_Norm_Conv2d(nn.Conv2d):

    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True):
        super(Pre_Norm_Conv2d, self).__init__(in_channels, out_channels,
                                              kernel_size, stride, padding,
                                              dilation, groups, bias)

    def forward(self, x):
        weight = self.weight
        weight_mean = weight.mean(dim=(1, 2, 3), keepdim=True)
        weight = weight - weight_mean
        std = weight.view(weight.size(0), -1).std(dim=1).view(-1, 1, 1, 1) + 1e-5
        weight = weight / std.expand_as(weight)
        return F.conv2d(x, weight, self.bias, self.stride,
                        self.padding, self.dilation, self.groups)


class GCI(nn.Module):
    """Global Cue Injection
    Takes shallow features with low receptive
    field and augments it with global info via
    adaptive instance normalisation"""

    def __init__(self, NormLayer=nn.BatchNorm2d):
        super(GCI, self).__init__()

        self.NormLayer = NormLayer
        self.from_scratch_layers = []

        self._init_params()

    def _conv2d(self, *args, **kwargs):
        conv = nn.Conv2d(*args, **kwargs)
        self.from_scratch_layers.append(conv)
        torch.nn.init.kaiming_normal_(conv.weight)
        return conv

    def _bnorm(self, *args, **kwargs):
        bn = self.NormLayer(*args, **kwargs)
        # self.bn_learn.append(bn)
        self.from_scratch_layers.append(bn)
        if bn.weight is not None:
            bn.weight.data.fill_(1)
            bn.bias.data.zero_()
        return bn

    def _init_params(self):

        self.fc_deep = nn.Sequential(Pre_Norm_Conv2d(256, 512, 1, bias=False),
                                     #  Pre_Norm_Conv2d(64, 512, 1, bias=False),
                                     #  self._bnorm(512),
                                     #  nn.BatchNorm2d(512, track_running_stats = False),
                                     nn.GroupNorm(32, 512),
                                     nn.ReLU())

        self.fc_skip = nn.Sequential(Pre_Norm_Conv2d(256, 256, 1, bias=False),
                                     #  Pre_Norm_Conv2d(64, 256, 1, bias=False),
                                     #  nn.BatchNorm2d(256, track_running_stats = False, affine=False),
                                     #  self._bnorm(256, affine=False)
                                     nn.GroupNorm(32, 256))

        self.fc_cls = nn.Sequential(Pre_Norm_Conv2d(256, 256, 1, bias=False),
                                    # Pre_Norm_Conv2d(256, 64, 1, bias=False),
                                    # nn.BatchNorm2d(256, track_running_stats = False),
                                    nn.GroupNorm(32, 256),
                                    # nn.GroupNorm(32,64),
                                    # self._bnorm(64),
                                    nn.ReLU())

    def forward(self, x, y):
        """Forward pass

        Args:
            x: shalow features
            y: deep features
        """

        # extract global attributes
        y = self.fc_deep(y)
        attrs, _ = y.view(y.size(0), y.size(1), -1).max(-1)

        # pre-process shallow features
        x = self.fc_skip(x)
        x = F.relu(self._adin_conv(x, attrs))

        return self.fc_cls(x)

    def _adin_conv(self, x, y):

        bs, num_c, _, _ = x.size()
        assert 2 * num_c == y.size(1), "AdIN: dimension mismatch"

        y = y.view(bs, 2, num_c)
        gamma, beta = y[:, 0], y[:, 1]

        gamma = gamma.unsqueeze(-1).unsqueeze(-1)
        beta = beta.unsqueeze(-1).unsqueeze(-1)

        return x * (gamma + 1) + beta



from einops import rearrange

import torch
import torch.nn as nn
import torch.nn.functional as F


class ResidualDiscritizer(nn.Module):
    def __init__(self, n_classes, feat_dim, norm='l2'):
        super(ResidualDiscritizer, self).__init__()
        self.norm = norm
        self.feat_dim = feat_dim
        self.n_classes = n_classes
        self.totems = nn.parameter.Parameter(torch.randn(feat_dim, n_classes), requires_grad=True)

    def forward(self, feature):
        """Computes the residual representation between features and a set of learnable classes.

        Args:
            feature (torch.tensor): A float tensor of shape [batch_size, n_frames, n_tokens, token_dim].

        Returns:
            torch.tensor: A float tensor of shape [batch_size, n_frames, n_tokens, n_classes].
        """
        if self.norm == 'l2':
            norm_vector = torch.linalg.vector_norm(feature, ord=2, dim=3, keepdim=True)
            feature = feature / norm_vector
        elif self.norm == 'l1':
            norm_vector = torch.linalg.vector_norm(feature, ord=1, dim=3, keepdim=True)
            feature = feature / norm_vector
        elif self.norm is None:
            pass
        else:
            raise NotImplementedError(f'Norm of type "{self.norm}" not implemented.')

        totems = self.totems.reshape(1, 1, 1, self.feat_dim, self.n_classes)
        residual_feats = (feature.unsqueeze(-1) - totems) / self.n_classes

        return residual_feats


class Gumbel_Softmax(nn.Module):
    def __init__(self, n_codewords, in_feat_dim, out_feat_dim=None, init_temperature=0.01, norm=None, **kwargs):
        super(Gumbel_Softmax, self).__init__()
        self.norm = norm
        self.n_codewords = n_codewords
        self.init_temperature = init_temperature

        if out_feat_dim is not None:
            self.proj = nn.Conv2d(in_feat_dim, out_feat_dim, kernel_size=1)
            self.codebook = nn.Embedding(n_codewords, out_feat_dim)
        else:
            self.proj = None
            self.codebook = nn.Embedding(n_codewords, in_feat_dim)
            out_feat_dim = in_feat_dim
        self.out_feat_dim = out_feat_dim

        self.codebook.weight.data.normal_()

        self.id_to_rgb = torch.rand(n_codewords, 3)

    def mat_id_to_rgb(self, codework_mask):
        # codework_mask: [height, width]
        return self.id_to_rgb[codework_mask]

    def forward(self, feats, temp=None):
        # feats: [batch_size, n_channels, height, width]

        if temp is None:
            temp = self.init_temperature

        # Condense number of channels
        if self.proj:
            feats = self.proj(feats)
        batch_size, n_channels, height, width = feats.shape

        # Compute difference between features and codebook words.
        # feats: [batch_size, n_channels, height, width]
        feats = feats.flatten(2).permute(0, 2, 1).unsqueeze(2)  # [batch_size, height*width, 1, n_channels]
        codebook_weights = self.codebook.weight.unsqueeze(0).unsqueeze(0)  # [1, 1, n_codewords, feat_dim]

        ## Compute L2 norm of features and codebook.
        if self.norm == 'l2':
            feats = F.normalize(feats, p=2.0, dim=3)
            codebook_weights = F.normalize(codebook_weights, p=2.0, dim=3)
        elif self.norm is None:
            pass
        else:
            raise NotImplementedError(f'Norm mode "{self.norm}" is not implemented for Gumbel Softmax class.')

        affinity = ((feats - codebook_weights).sum(dim=-1))**2

        # Compute logits.
        # [batch_size, n_feats, n_classes, feat_dim]
        logits = torch.exp(-affinity)

        # Compute Gumbel-Softmax.
        # [batch_size, n_feats, n_classes, feat_dim]
        soft_one_hot = F.gumbel_softmax(logits, tau=temp, dim=2, hard=False)

        # Compute KL divergence.
        qy = F.softmax(logits, dim=2)

        # V1: Taming Transformers Gumbel; https://github.com/CompVis/taming-transformers/blob/24268930bf1dce879235a7fddd0b2355b84d7ea6/taming/modules/vqvae/quantize.py#L191
        kld_loss = torch.sum(qy * torch.log(qy * self.n_codewords + 1e-10), dim=2).mean()

        # V2: Dalle; https://github.com/lucidrains/DALLE-pytorch/blob/289fddbef7e5a9b4c6f3f223a7505c34afe4ac67/dalle_pytorch/dalle_pytorch.py#L245
        # NOTE: Pushes optimization to uniform distribution and model does not learn.
        # log_uniform = torch.ones(self.n_codewords, device=qy.device) / self.n_codewords
        # kld_loss = F.kl_div(log_uniform.log(), qy.log(), None, None, 'batchmean', log_target=True)

        # V3: My idea to encourage confident predictions.
        # kld_loss = F.kl_div(qy, soft_one_hot, size_average=None, reduce=None, reduction='batchmean', log_target=True)

        # breakpoint()

        # Sample codebook for codewords.
        # [batch_size, n_feats, feat_dim]
        codewords = soft_one_hot.argmax(dim=2)
        sel_codewords = self.codebook(codewords)

        # Resize output features.
        sel_codewords = torch.matmul(soft_one_hot,
                                     self.codebook.weight).permute(0, 2, 1).reshape(batch_size, n_channels, height,
                                                                                    width)
        codewords = codewords.reshape([batch_size, height, width])
        mat_id_probs = soft_one_hot.permute(0, 2, 1).reshape(batch_size, self.n_codewords, height, width)

        return sel_codewords, codewords, kld_loss, mat_id_probs


class VectorQuantizer2(nn.Module):
    def __init__(self, n_codewords, in_feat_dim, out_feat_dim=None, init_temperature=0.01, norm=None, **kwargs):
        super(VectorQuantizer2, self).__init__()
        self.beta = kwargs['beta']
        self.norm = norm
        self.n_codewords = n_codewords
        self.init_temperature = init_temperature

        if out_feat_dim is not None:
            self.proj = nn.Conv2d(in_feat_dim, out_feat_dim, kernel_size=1)
            self.codebook = nn.Embedding(n_codewords, out_feat_dim)
            codeword_dim = out_feat_dim
        else:
            self.proj = None
            self.codebook = nn.Embedding(n_codewords, in_feat_dim)
            codeword_dim = in_feat_dim
            out_feat_dim = in_feat_dim
        self.out_feat_dim = out_feat_dim

        self.codebook.weight.data.uniform_(-1.0 / n_codewords, 1.0 / codeword_dim)

        self.id_to_rgb = torch.rand(n_codewords, 3)

    def mat_id_to_rgb(self, codework_mask):
        # codework_mask: [height, width]
        return self.id_to_rgb[codework_mask]

    def mat_id_to_mat_feat(self, mat_id_mask):
        # mat_id_mask: [height, width]
        return self.codebook(mat_id_mask)

    def forward(self, feats, temp=None):
        # feats: [batch_size, n_channels, height, width]

        if temp is None:
            temp = self.init_temperature

        # Condense number of channels
        if self.proj:
            feats = self.proj(feats)

        batch_size, n_channels, height, width = feats.shape
        z = feats.permute(0, 2, 3, 1).contiguous()  # [batch_size, height, width, feat_dim]
        zf = z.view(-1, n_channels)

        d = torch.sum(zf ** 2, dim=1, keepdim=True) + \
            torch.sum(self.codebook.weight**2, dim=1) - 2 * \
            torch.einsum('bd,dn->bn', zf, rearrange(self.codebook.weight, 'n d -> d n'))

        min_encoding_indices = torch.argmin(d, dim=1)
        mat_id_probs = F.softmax(-torch.abs(d), dim=1)
        z_q = self.codebook(min_encoding_indices).view(z.shape)

        # compute loss for embedding
        loss = self.beta * torch.mean((z_q.detach() - z)**2) + torch.mean((z_q - z.detach())**2)

        # preserve gradients
        z_q = z + (z_q - z).detach()

        # reshape back to match original input shape
        z_q = rearrange(z_q, 'b h w c -> b c h w').contiguous()
        min_encoding_indices = min_encoding_indices.reshape(batch_size, height, width)
        mat_id_probs = mat_id_probs.reshape(batch_size, height, width, self.n_codewords).permute(0, 3, 1, 2)

        return z_q, min_encoding_indices, loss, mat_id_probs


if __name__ == '__main__':
    batch_size, n_frames, n_tokens, token_dim = 4, 5, 25, 100
    test_feats = torch.randn([batch_size, n_frames, n_tokens, token_dim])

    n_classes = 30
    norm = 'l2'
    discritizer = ResidualDiscritizer(n_classes, token_dim, norm=norm)

    output_feats = discritizer(test_feats)



import ubelt as ub


def test_init_from_pretrained_state():
    # Train a demo model (in the future grab a pretrained demo model)
    import ubelt as ub
    from watch.tasks.fusion.fit import fit_model  # NOQA
    import kwcoco

    import pytest
    pytest.skip('slow test')

    # import xdev
    # xdev.make_warnings_print_tracebacks()

    # args = None
    # cmdline = False
    gpus = None
    test_dpath = ub.Path.appdir('watch/test/fusion/').ensuredir()
    results_path = ub.ensuredir((test_dpath, 'predict'))
    ub.delete(results_path)
    ub.ensuredir(results_path)
    package_fpath = test_dpath / 'my_test_package.pt'
    # train_dset = kwcoco.CocoDataset.demo('special:vidshapes4-multispectral', num_frames=5, gsize=(128, 128))
    test_dset = kwcoco.CocoDataset.demo('special:vidshapes2-multispectral', num_frames=3, gsize=(128, 128))
    fit_kwargs = {
         'train_dataset': test_dset.fpath,
         'datamodule': 'KWCocoVideoDataModule',
         'workdir': ub.ensuredir((test_dpath, 'train')),
         'package_fpath': package_fpath,
         'max_epochs': 1,
         'time_steps': 3,
         'chip_size': 64,
         'max_steps': 1,
         'learning_rate': 1e-5,
         'diff_inputs': False,
         'num_workers': 1,
         'gpus': gpus,
    }
    package_fpath = fit_model(**fit_kwargs)

    # Start a new training run but try to init from the pretrianed state
    # (even though it is different)
    fit_kwargs2 = fit_kwargs.copy()
    fit_kwargs2['init'] = package_fpath
    fit_kwargs2['diff_inputs'] = 0
    fit_model(**fit_kwargs2)

    # Start a new training run but try to init from the pretrianed state
    # (even though it is different)  # TODO: had to remove diff-inputs, make more different to test
    fit_kwargs3 = fit_kwargs.copy()
    fit_kwargs3['init'] = 'kaiming_normal'
    fit_kwargs3['diff_inputs'] = 0
    fit_model(**fit_kwargs2)


def test_init_from_phase1_models():
    import watch
    import os
    try:
        phase1_dvc_dpath = watch.find_dvc_dpath(tags='phase1')
    except Exception:
        import pytest
        pytest.skip('dvc repo is not available')

    from watch.tasks.fusion import production
    found = None
    for row in production.PRODUCTION_MODELS:
        if row['name'] == 'Drop3_SpotCheck_V323_epoch=18-step=12976.pt':
            found = row
            break
    assert found
    model_fpath = phase1_dvc_dpath / found['file_name']
    if not model_fpath.exists():
        import pytest
        pytest.skip('dvc repo is not available')

    init = model_fpath

    # First test it works by itself
    from watch.tasks.fusion.fit import coerce_initializer
    initializer = coerce_initializer(init)
    fpath = initializer._rectify_fpath()
    from torch_liberator.initializer import _torch_load
    with open(fpath, 'rb') as file:
        state_dict = _torch_load(file)
    assert state_dict

    # gpus = None
    test_dpath = ub.Path.appdir('watch/test/fusion/').ensuredir()
    results_path = ub.ensuredir((test_dpath, 'predict'))
    ub.delete(results_path)
    ub.ensuredir(results_path)
    package_fpath = test_dpath / 'my_test_package.pt'
    import kwcoco
    from watch.tasks.fusion.fit import fit_model  # NOQA
    test_dset = kwcoco.CocoDataset.demo('special:vidshapes2-multispectral', num_frames=3, gsize=(128, 128))
    fit_kwargs = {
         'train_dataset': test_dset.fpath,
         'datamodule': 'KWCocoVideoDataModule',
         'workdir': ub.ensuredir((test_dpath, 'train')),
         'package_fpath': os.fspath(package_fpath),
         'max_epochs': 1,
         'time_steps': 3,
         'chip_size': 64,
         'max_steps': 1,
         'learning_rate': 1e-5,
         'diff_inputs': False,
         'num_workers': 1,
         'init': os.fspath(model_fpath),
         'devices': None,
         # 'gpus': gpus,
    }
    package_fpath = fit_model(**fit_kwargs)


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/watch/tests/test_init_from_pretrained.py
    """
    test_init_from_pretrained_state()
